"""Revocation through ledger agnostic AnonCreds interface."""

import asyncio
import hashlib
import http
import json
import logging
import os
import time
from pathlib import Path
from typing import List, Mapping, NamedTuple, Optional, Sequence, Tuple, Union
from urllib.parse import urlparse

import base58
from anoncreds import (
    AnoncredsError,
    Credential,
    CredentialRevocationConfig,
    RevocationRegistryDefinition,
    RevocationRegistryDefinitionPrivate,
    RevocationStatusList,
    W3cCredential,
)
from aries_askar import Entry
from aries_askar.error import AskarError
from requests import RequestException, Session
from uuid_utils import uuid4

from ..askar.profile_anon import AskarAnonCredsProfile, AskarAnonCredsProfileSession
from ..core.error import BaseError
from ..core.event_bus import Event, EventBus
from ..core.profile import Profile, ProfileSession
from ..tails.anoncreds_tails_server import AnonCredsTailsServer
from .error_messages import ANONCREDS_PROFILE_REQUIRED_MSG
from .events import RevListFinishedEvent, RevRegDefFinishedEvent
from .issuer import (
    CATEGORY_CRED_DEF,
    CATEGORY_CRED_DEF_PRIVATE,
    STATE_FINISHED,
    AnonCredsIssuer,
)
from .models.credential_definition import CredDef
from .models.revocation import (
    RevList,
    RevListResult,
    RevListState,
    RevRegDef,
    RevRegDefResult,
    RevRegDefState,
)
from .registry import AnonCredsRegistry
from .util import indy_client_dir

LOGGER = logging.getLogger(__name__)

CATEGORY_REV_LIST = "revocation_list"
CATEGORY_REV_REG_DEF = "revocation_reg_def"
CATEGORY_REV_REG_DEF_PRIVATE = "revocation_reg_def_private"
CATEGORY_REV_REG_ISSUER = "revocation_reg_def_issuer"
STATE_REVOCATION_POSTED = "posted"
STATE_REVOCATION_PENDING = "pending"
REV_REG_DEF_STATE_ACTIVE = "active"


class AnonCredsRevocationError(BaseError):
    """Generic revocation error."""


class AnonCredsRevocationRegistryFullError(AnonCredsRevocationError):
    """Revocation registry is full when issuing a new credential."""


class RevokeResult(NamedTuple):
    """RevokeResult."""

    prev: RevList
    curr: Optional[RevList] = None
    revoked: Optional[Sequence[int]] = None
    failed: Optional[Sequence[str]] = None


class AnonCredsRevocation:
    """Revocation registry operations manager."""

    def __init__(self, profile: Profile):
        """Initialize an AnonCredsRevocation instance.

        Args:
            profile: The active profile instance

        """
        self._profile = profile

    @property
    def profile(self) -> AskarAnonCredsProfile:
        """Accessor for the profile instance."""
        if not isinstance(self._profile, AskarAnonCredsProfile):
            raise ValueError(ANONCREDS_PROFILE_REQUIRED_MSG)

        return self._profile

    async def notify(self, event: Event) -> None:
        """Emit an event on the event bus."""
        event_bus = self.profile.inject(EventBus)
        await event_bus.notify(self.profile, event)

    async def _finish_registration(
        self,
        txn: AskarAnonCredsProfileSession,
        category: str,
        job_id: str,
        registered_id: str,
        *,
        state: Optional[str] = None,
    ):
        entry = await txn.handle.fetch(
            category,
            job_id,
            for_update=True,
        )
        if not entry:
            raise AnonCredsRevocationError(
                f"{category} with job id {job_id} could not be found"
            )

        if state:
            tags = entry.tags
            tags["state"] = state
        else:
            tags = entry.tags

        await txn.handle.insert(
            category,
            registered_id,
            value=entry.value,
            tags=tags,
        )
        await txn.handle.remove(category, job_id)
        return entry

    async def create_and_register_revocation_registry_definition(
        self,
        issuer_id: str,
        cred_def_id: str,
        registry_type: str,
        tag: str,
        max_cred_num: int,
        options: Optional[dict] = None,
    ) -> RevRegDefResult:
        """Create a new revocation registry and register on network.

        Args:
            issuer_id (str): issuer identifier
            cred_def_id (str): credential definition identifier
            registry_type (str): revocation registry type
            tag (str): revocation registry tag
            max_cred_num (int): maximum number of credentials supported
            options (dict): revocation registry options

        Returns:
            RevRegDefResult: revocation registry definition result

        """
        options = options or {}
        try:
            async with self.profile.session() as session:
                cred_def = await session.handle.fetch(CATEGORY_CRED_DEF, cred_def_id)
        except AskarError as err:
            raise AnonCredsRevocationError(
                "Error retrieving credential definition"
            ) from err

        if not cred_def:
            raise AnonCredsRevocationError(
                "Credential definition not found for revocation registry"
            )

        tails_dir = indy_client_dir("tails", create=True)

        try:
            (
                rev_reg_def,
                rev_reg_def_private,
            ) = await asyncio.get_event_loop().run_in_executor(
                None,
                lambda: RevocationRegistryDefinition.create(
                    cred_def_id,
                    cred_def.raw_value,
                    issuer_id,
                    tag,
                    registry_type,
                    max_cred_num,
                    tails_dir_path=tails_dir,
                ),
            )
        except AnoncredsError as err:
            raise AnonCredsRevocationError("Error creating revocation registry") from err

        rev_reg_def = RevRegDef.from_native(rev_reg_def)

        public_tails_uri = self.generate_public_tails_uri(rev_reg_def)
        rev_reg_def.value.tails_location = public_tails_uri

        anoncreds_registry = self.profile.inject(AnonCredsRegistry)
        result = await anoncreds_registry.register_revocation_registry_definition(
            self.profile, rev_reg_def, options
        )

        await self.store_revocation_registry_definition(
            result, rev_reg_def_private, options
        )
        return result

    async def store_revocation_registry_definition(
        self,
        result: RevRegDefResult,
        rev_reg_def_private: RevocationRegistryDefinitionPrivate,
        options: Optional[dict] = None,
    ) -> None:
        """Store a revocation registry definition."""
        options = options or {}
        identifier = result.job_id or result.rev_reg_def_id
        if not identifier:
            raise AnonCredsRevocationError(
                "Revocation registry definition id or job id not found"
            )

        rev_reg_def = (
            result.revocation_registry_definition_state.revocation_registry_definition
        )

        try:
            async with self.profile.transaction() as txn:
                await txn.handle.insert(
                    CATEGORY_REV_REG_DEF,
                    identifier,
                    rev_reg_def.to_json(),
                    tags={
                        "cred_def_id": rev_reg_def.cred_def_id,
                        "state": result.revocation_registry_definition_state.state,
                        "active": json.dumps(False),
                    },
                )
                await txn.handle.insert(
                    CATEGORY_REV_REG_DEF_PRIVATE,
                    identifier,
                    rev_reg_def_private.to_json_buffer(),
                )
                await txn.commit()

            if result.revocation_registry_definition_state.state == STATE_FINISHED:
                await self.notify(
                    RevRegDefFinishedEvent.with_payload(identifier, rev_reg_def, options)
                )
        except AskarError as err:
            raise AnonCredsRevocationError(
                "Error saving new revocation registry"
            ) from err

    async def finish_revocation_registry_definition(
        self, job_id: str, rev_reg_def_id: str, options: Optional[dict] = None
    ) -> None:
        """Mark a rev reg def as finished."""
        options = options or {}
        async with self.profile.transaction() as txn:
            entry = await self._finish_registration(
                txn, CATEGORY_REV_REG_DEF, job_id, rev_reg_def_id, state=STATE_FINISHED
            )
            rev_reg_def = RevRegDef.from_json(entry.value)
            await self._finish_registration(
                txn,
                CATEGORY_REV_REG_DEF_PRIVATE,
                job_id,
                rev_reg_def_id,
            )
            await txn.commit()

        await self.notify(
            RevRegDefFinishedEvent.with_payload(rev_reg_def_id, rev_reg_def, options)
        )

    async def get_created_revocation_registry_definitions(
        self,
        cred_def_id: Optional[str] = None,
        state: Optional[str] = None,
    ) -> Sequence[str]:
        """Retrieve IDs of rev reg defs previously created."""
        async with self.profile.session() as session:
            # TODO limit? scan?
            rev_reg_defs = await session.handle.fetch_all(
                CATEGORY_REV_REG_DEF,
                {
                    key: value
                    for key, value in {
                        "cred_def_id": cred_def_id,
                        "state": state,
                    }.items()
                    if value is not None
                },
            )
        # entry.name was stored as the credential_definition's ID
        return [entry.name for entry in rev_reg_defs]

    async def get_created_revocation_registry_definition_state(
        self,
        rev_reg_def_id: str,
    ) -> Optional[str]:
        """Retrieve rev reg def by ID from rev reg defs previously created."""
        async with self.profile.session() as session:
            rev_reg_def_entry = await session.handle.fetch(
                CATEGORY_REV_REG_DEF,
                name=rev_reg_def_id,
            )

        if rev_reg_def_entry:
            return rev_reg_def_entry.tags.get("state")

        return None

    async def get_created_revocation_registry_definition(
        self,
        rev_reg_def_id: str,
    ) -> Optional[RevRegDef]:
        """Retrieve rev reg def by ID from rev reg defs previously created."""
        async with self.profile.session() as session:
            rev_reg_def_entry = await session.handle.fetch(
                CATEGORY_REV_REG_DEF,
                name=rev_reg_def_id,
            )

        if rev_reg_def_entry:
            return RevRegDef.deserialize(rev_reg_def_entry.value_json)

        return None

    async def set_active_registry(self, rev_reg_def_id: str) -> None:
        """Mark a registry as active."""
        async with self.profile.transaction() as txn:
            entry = await txn.handle.fetch(
                CATEGORY_REV_REG_DEF,
                rev_reg_def_id,
                for_update=True,
            )
            if not entry:
                raise AnonCredsRevocationError(
                    f"{CATEGORY_REV_REG_DEF} with id {rev_reg_def_id} could not be found"
                )

            if entry.tags["active"] == json.dumps(True):
                # NOTE If there are other registries set as active, we're not
                # clearing them if the one we want to be active is already
                # active. This probably isn't an issue.
                return

            cred_def_id = entry.tags["cred_def_id"]

            old_active_entries = await txn.handle.fetch_all(
                CATEGORY_REV_REG_DEF,
                {
                    "active": json.dumps(True),
                    "cred_def_id": cred_def_id,
                },
                for_update=True,
            )

            if len(old_active_entries) > 1:
                LOGGER.error(
                    "More than one registry was set as active for "
                    f"cred def {cred_def_id}; clearing active tag from all records"
                )

            for old_entry in old_active_entries:
                tags = old_entry.tags
                tags["active"] = json.dumps(False)
                await txn.handle.replace(
                    CATEGORY_REV_REG_DEF,
                    old_entry.name,
                    old_entry.value,
                    tags,
                )

            tags = entry.tags
            tags["active"] = json.dumps(True)
            await txn.handle.replace(
                CATEGORY_REV_REG_DEF,
                rev_reg_def_id,
                value=entry.value,
                tags=tags,
            )
            await txn.commit()

    async def create_and_register_revocation_list(
        self, rev_reg_def_id: str, options: Optional[dict] = None
    ) -> RevListResult:
        """Create and register a revocation list."""
        options = options or {}
        try:
            async with self.profile.session() as session:
                rev_reg_def_entry = await session.handle.fetch(
                    CATEGORY_REV_REG_DEF, rev_reg_def_id
                )
                rev_reg_def_private_entry = await session.handle.fetch(
                    CATEGORY_REV_REG_DEF_PRIVATE, rev_reg_def_id
                )
        except AskarError as err:
            raise AnonCredsRevocationError(
                "Error retrieving required revocation registry definition data"
            ) from err

        if not rev_reg_def_entry or not rev_reg_def_private_entry:
            raise AnonCredsRevocationError(
                (
                    "Missing required revocation registry data: "
                    "revocation registry definition"
                    if not rev_reg_def_entry
                    else ""
                ),
                (
                    "revocation registry private definition"
                    if not rev_reg_def_private_entry
                    else ""
                ),
            )

        try:
            async with self.profile.session() as session:
                cred_def_entry = await session.handle.fetch(
                    CATEGORY_CRED_DEF, rev_reg_def_entry.value_json["credDefId"]
                )
        except AskarError as err:
            raise AnonCredsRevocationError(
                f"Error retrieving cred def {rev_reg_def_entry.value_json['credDefId']}"
            ) from err

        rev_reg_def = RevRegDef.deserialize(rev_reg_def_entry.value_json)
        cred_def = CredDef.deserialize(cred_def_entry.value_json)
        rev_reg_def_private = RevocationRegistryDefinitionPrivate.load(
            rev_reg_def_private_entry.value_json
        )
        # TODO This is a little rough; stored tails location will have public uri
        rev_reg_def.value.tails_location = self.get_local_tails_path(rev_reg_def)

        rev_list = RevocationStatusList.create(
            cred_def.to_native(),
            rev_reg_def_id,
            rev_reg_def.to_native(),
            rev_reg_def_private,
            rev_reg_def.issuer_id,
        )

        anoncreds_registry = self.profile.inject(AnonCredsRegistry)
        result = await anoncreds_registry.register_revocation_list(
            self.profile, rev_reg_def, RevList.from_native(rev_list), options
        )

        if options.get("failed_to_upload", False):
            result.revocation_list_state.state = RevListState.STATE_FAILED

        await self.store_revocation_registry_list(result)

        return result

    async def store_revocation_registry_list(self, result: RevListResult) -> None:
        """Store a revocation registry list."""

        identifier = result.job_id or result.rev_reg_def_id
        if not identifier:
            raise AnonCredsRevocationError(
                "Revocation registry definition id or job id not found"
            )

        rev_list = result.revocation_list_state.revocation_list
        try:
            async with self.profile.session() as session:
                await session.handle.insert(
                    CATEGORY_REV_LIST,
                    identifier,
                    value_json={
                        "rev_list": rev_list.serialize(),
                        # AnonCreds uses the 0 index internally
                        # and can't be used for a credential
                        "next_index": 1,
                        "pending": None,
                    },
                    tags={
                        "state": result.revocation_list_state.state,
                        "pending": json.dumps(False),
                    },
                )

            if result.revocation_list_state.state == STATE_FINISHED:
                await self.notify(
                    RevListFinishedEvent.with_payload(
                        rev_list.rev_reg_def_id, rev_list.revocation_list
                    )
                )

        except AskarError as err:
            raise AnonCredsRevocationError(
                "Error saving new revocation registry"
            ) from err

    async def finish_revocation_list(
        self, job_id: str, rev_reg_def_id: str, revoked: list
    ) -> None:
        """Mark a revocation list as finished."""
        async with self.profile.transaction() as txn:
            # Finish the registration if the list is new, otherwise already updated
            existing_list = await txn.handle.fetch(
                CATEGORY_REV_LIST,
                rev_reg_def_id,
            )
            if not existing_list:
                await self._finish_registration(
                    txn,
                    CATEGORY_REV_LIST,
                    job_id,
                    rev_reg_def_id,
                    state=STATE_FINISHED,
                )
                await txn.commit()
            # Notify about revoked creds on any list update
            await self.notify(RevListFinishedEvent.with_payload(rev_reg_def_id, revoked))

    async def update_revocation_list(
        self,
        rev_reg_def_id: str,
        prev: RevList,
        curr: RevList,
        revoked: Sequence[int],
        options: Optional[dict] = None,
    ) -> RevListResult:
        """Publish and update to a revocation list."""
        options = options or {}
        try:
            async with self.profile.session() as session:
                rev_reg_def_entry = await session.handle.fetch(
                    CATEGORY_REV_REG_DEF, rev_reg_def_id
                )
        except AskarError as err:
            raise AnonCredsRevocationError(
                "Error retrieving revocation registry definition"
            ) from err

        if not rev_reg_def_entry:
            raise AnonCredsRevocationError(
                f"Revocation registry definition not found for id {rev_reg_def_id}"
            )

        try:
            async with self.profile.session() as session:
                rev_list_entry = await session.handle.fetch(
                    CATEGORY_REV_LIST, rev_reg_def_id
                )
        except AskarError as err:
            raise AnonCredsRevocationError("Error retrieving revocation list") from err

        if not rev_list_entry:
            raise AnonCredsRevocationError(
                f"Revocation list not found for id {rev_reg_def_id}"
            )

        rev_reg_def = RevRegDef.deserialize(rev_reg_def_entry.value_json)
        rev_list = RevList.deserialize(rev_list_entry.value_json["rev_list"])
        if rev_list.revocation_list != curr.revocation_list:
            raise AnonCredsRevocationError("Passed revocation list does not match stored")

        anoncreds_registry = self.profile.inject(AnonCredsRegistry)
        result = await anoncreds_registry.update_revocation_list(
            self.profile, rev_reg_def, prev, curr, revoked, options
        )

        try:
            async with self.profile.session() as session:
                rev_list_entry_upd = await session.handle.fetch(
                    CATEGORY_REV_LIST, result.rev_reg_def_id, for_update=True
                )
                if not rev_list_entry_upd:
                    raise AnonCredsRevocationError(
                        f"Revocation list not found for id {rev_reg_def_id}"
                    )
                tags = rev_list_entry_upd.tags
                tags["state"] = result.revocation_list_state.state
                await session.handle.replace(
                    CATEGORY_REV_LIST,
                    result.rev_reg_def_id,
                    value=rev_list_entry_upd.value,
                    tags=tags,
                )
        except AskarError as err:
            raise AnonCredsRevocationError(
                "Error saving new revocation registry"
            ) from err

        return result

    async def get_created_revocation_list(self, rev_reg_def_id: str) -> Optional[RevList]:
        """Return rev list from record in wallet."""
        try:
            async with self.profile.session() as session:
                rev_list_entry = await session.handle.fetch(
                    CATEGORY_REV_LIST, rev_reg_def_id
                )
        except AskarError as err:
            raise AnonCredsRevocationError("Error retrieving revocation list") from err

        if rev_list_entry:
            return RevList.deserialize(rev_list_entry.value_json["rev_list"])

        return None

    async def get_revocation_lists_with_pending_revocations(self) -> Sequence[str]:
        """Return a list of rev reg def ids with pending revocations."""
        try:
            async with self.profile.session() as session:
                rev_list_entries = await session.handle.fetch_all(
                    CATEGORY_REV_LIST,
                    {"pending": json.dumps(True)},
                )
        except AskarError as err:
            raise AnonCredsRevocationError("Error retrieving revocation list") from err

        if rev_list_entries:
            return [entry.name for entry in rev_list_entries]

        return []

    async def retrieve_tails(self, rev_reg_def: RevRegDef) -> str:
        """Retrieve tails file from server."""
        LOGGER.info(
            "Downloading the tails file with hash: %s",
            rev_reg_def.value.tails_hash,
        )

        tails_file_path = Path(self.get_local_tails_path(rev_reg_def))
        tails_file_dir = tails_file_path.parent
        if not tails_file_dir.exists():
            tails_file_dir.mkdir(parents=True)

        buffer_size = 65536  # should be multiple of 32 bytes for sha256
        file_hasher = hashlib.sha256()
        with open(tails_file_path, "wb", buffer_size) as tails_file:
            with Session() as req_session:
                try:
                    resp = req_session.get(rev_reg_def.value.tails_location, stream=True)
                    # Should this directly raise an Error?
                    if resp.status_code != http.HTTPStatus.OK:
                        LOGGER.warning(
                            f"Unexpected status code for tails file: {resp.status_code}"
                        )
                    for buf in resp.iter_content(chunk_size=buffer_size):
                        tails_file.write(buf)
                        file_hasher.update(buf)
                except RequestException as rx:
                    raise AnonCredsRevocationError(f"Error retrieving tails file: {rx}")

        download_tails_hash = base58.b58encode(file_hasher.digest()).decode("utf-8")
        if download_tails_hash != rev_reg_def.value.tails_hash:
            try:
                os.remove(tails_file_path)
            except OSError as err:
                LOGGER.warning(f"Could not delete invalid tails file: {err}")

            raise AnonCredsRevocationError(
                "The hash of the downloaded tails file does not match."
            )

        return str(tails_file_path)

    def _check_url(self, url: str) -> None:
        parsed = urlparse(url)
        if not (parsed.scheme and parsed.netloc and parsed.path):
            raise AnonCredsRevocationError("URI {} is not a valid URL".format(url))

    def generate_public_tails_uri(self, rev_reg_def: RevRegDef) -> str:
        """Construct tails uri from rev_reg_def."""
        tails_base_url = self.profile.settings.get("tails_server_base_url")
        if not tails_base_url:
            raise AnonCredsRevocationError("tails_server_base_url not configured")

        public_tails_uri = (
            tails_base_url.rstrip("/") + f"/hash/{rev_reg_def.value.tails_hash}"
        )

        self._check_url(public_tails_uri)
        return public_tails_uri

    def get_local_tails_path(self, rev_reg_def: RevRegDef) -> str:
        """Get the local path to the tails file."""
        tails_dir = indy_client_dir("tails", create=False)
        return os.path.join(tails_dir, rev_reg_def.value.tails_hash)

    async def upload_tails_file(self, rev_reg_def: RevRegDef) -> None:
        """Upload the local tails file to the tails server."""
        tails_server = AnonCredsTailsServer()

        if not Path(self.get_local_tails_path(rev_reg_def)).is_file():
            raise AnonCredsRevocationError("Local tails file not found")

        (upload_success, result) = await tails_server.upload_tails_file(
            self.profile.context,
            rev_reg_def.value.tails_hash,
            self.get_local_tails_path(rev_reg_def),
            interval=0.8,
            backoff=-0.5,
            max_attempts=5,  # heuristic: respect HTTP timeout
        )

        if not upload_success:
            raise AnonCredsRevocationError(
                f"Tails file for rev reg for {rev_reg_def.cred_def_id} "
                f"failed to upload: {result}"
            )
        if rev_reg_def.value.tails_location != result:
            raise AnonCredsRevocationError(
                f"Tails file for rev reg for {rev_reg_def.cred_def_id} "
                f"uploaded to wrong location: {result} "
                f"(should have been {rev_reg_def.value.tails_location})"
            )

    async def get_or_fetch_local_tails_path(self, rev_reg_def: RevRegDef) -> str:
        """Return path to local tails file.

        If not present, retrieve from tails server.
        """
        tails_file_path = self.get_local_tails_path(rev_reg_def)
        if Path(tails_file_path).is_file():
            return tails_file_path
        return await self.retrieve_tails(rev_reg_def)

    # Registry Management

    async def handle_full_registry(self, rev_reg_def_id: str) -> None:
        """Update the registry status and start the next registry generation."""
        async with self.profile.session() as session:
            active_rev_reg_def = await session.handle.fetch(
                CATEGORY_REV_REG_DEF, rev_reg_def_id
            )
            if active_rev_reg_def:
                # ok, we have an active rev reg.
                # find the backup/fallover rev reg (finished and not active)
                rev_reg_defs = await session.handle.fetch_all(
                    CATEGORY_REV_REG_DEF,
                    {
                        "active": json.dumps(False),
                        "cred_def_id": active_rev_reg_def.value_json["credDefId"],
                        "state": RevRegDefState.STATE_FINISHED,
                    },
                    limit=1,
                )
                if len(rev_reg_defs):
                    backup_rev_reg_def_id = rev_reg_defs[0].name
                else:
                    # attempted to create and register here but fails in practical usage.
                    # the indexes and list do not get set properly (timing issue?)
                    # if max cred num = 4 for instance, will get
                    # Revocation status list does not have the index 4
                    # in _create_credential calling Credential.create
                    raise AnonCredsRevocationError(
                        "Error handling full registry. No backup registry available."
                    )

        # set the backup to active...
        if backup_rev_reg_def_id:
            await self.set_active_registry(backup_rev_reg_def_id)

            async with self.profile.transaction() as txn:
                # re-fetch the old active (it's been updated), we need to mark as full
                active_rev_reg_def = await txn.handle.fetch(
                    CATEGORY_REV_REG_DEF, rev_reg_def_id, for_update=True
                )
                tags = active_rev_reg_def.tags
                tags["state"] = RevRegDefState.STATE_FULL
                await txn.handle.replace(
                    CATEGORY_REV_REG_DEF,
                    active_rev_reg_def.name,
                    active_rev_reg_def.value,
                    tags,
                )
                await txn.commit()

            # create our next fallover/backup
            backup_reg = await self.create_and_register_revocation_registry_definition(
                issuer_id=active_rev_reg_def.value_json["issuerId"],
                cred_def_id=active_rev_reg_def.value_json["credDefId"],
                registry_type=active_rev_reg_def.value_json["revocDefType"],
                tag=str(uuid4()),
                max_cred_num=active_rev_reg_def.value_json["value"]["maxCredNum"],
            )
            LOGGER.info(f"Previous rev_reg_def_id = {rev_reg_def_id}")
            LOGGER.info(f"Current rev_reg_def_id = {backup_rev_reg_def_id}")
            LOGGER.info(f"Backup reg = {backup_reg.rev_reg_def_id}")

    async def decommission_registry(self, cred_def_id: str) -> list:
        """Decommission post-init registries and start the next registry generation."""
        active_reg = await self.get_or_create_active_registry(cred_def_id)

        # create new one and set active
        new_reg = await self.create_and_register_revocation_registry_definition(
            issuer_id=active_reg.rev_reg_def.issuer_id,
            cred_def_id=active_reg.rev_reg_def.cred_def_id,
            registry_type=active_reg.rev_reg_def.type,
            tag=str(uuid4()),
            max_cred_num=active_reg.rev_reg_def.value.max_cred_num,
        )
        # set new as active...
        await self.set_active_registry(new_reg.rev_reg_def_id)

        # decommission everything except init/wait
        async with self.profile.transaction() as txn:
            registries = await txn.handle.fetch_all(
                CATEGORY_REV_REG_DEF,
                {
                    "cred_def_id": cred_def_id,
                },
                for_update=True,
            )

            recs = list(
                filter(
                    lambda r: r.tags.get("state") != RevRegDefState.STATE_WAIT,
                    registries,
                )
            )
            for rec in recs:
                if rec.name != new_reg.rev_reg_def_id:
                    tags = rec.tags
                    tags["active"] = json.dumps(False)
                    tags["state"] = RevRegDefState.STATE_DECOMMISSIONED
                    await txn.handle.replace(
                        CATEGORY_REV_REG_DEF,
                        rec.name,
                        rec.value,
                        tags,
                    )
            await txn.commit()
        # create a second one for backup, don't make it active
        backup_reg = await self.create_and_register_revocation_registry_definition(
            issuer_id=active_reg.rev_reg_def.issuer_id,
            cred_def_id=active_reg.rev_reg_def.cred_def_id,
            registry_type=active_reg.rev_reg_def.type,
            tag=str(uuid4()),
            max_cred_num=active_reg.rev_reg_def.value.max_cred_num,
        )

        LOGGER.info(f"New registry = {new_reg}")
        LOGGER.info(f"Backup registry = {backup_reg}")
        LOGGER.debug(f"Decommissioned registries = {recs}")
        return recs

    async def get_or_create_active_registry(self, cred_def_id: str) -> RevRegDefResult:
        """Get or create a revocation registry for the given cred def id."""
        async with self.profile.session() as session:
            rev_reg_defs = await session.handle.fetch_all(
                CATEGORY_REV_REG_DEF,
                {
                    "cred_def_id": cred_def_id,
                    "active": json.dumps(True),
                },
                limit=1,
            )

        if not rev_reg_defs:
            raise AnonCredsRevocationError("No active registry")

        entry = rev_reg_defs[0]

        rev_reg_def = RevRegDef.deserialize(entry.value_json)
        result = RevRegDefResult(
            None,
            RevRegDefState(
                state=STATE_FINISHED,
                revocation_registry_definition_id=entry.name,
                revocation_registry_definition=rev_reg_def,
            ),
            registration_metadata={},
            revocation_registry_definition_metadata={},
        )
        return result

    # Credential Operations

    async def create_credential_w3c(
        self,
        w3c_credential_offer: dict,
        w3c_credential_request: dict,
        w3c_credential_values: dict,
        *,
        retries: int = 5,
    ) -> Tuple[str, str, str]:
        """Create a w3c_credential.

        Args:
            w3c_credential_offer: Credential Offer to create w3c_credential for
            w3c_credential_request: Credential request to create w3c_credential for
            w3c_credential_values: Values to go in w3c_credential
            retries: number of times to retry w3c_credential creation

        Returns:
            A tuple of created w3c_credential and revocation id

        """
        return await self._create_credential_helper(
            w3c_credential_offer,
            w3c_credential_request,
            w3c_credential_values,
            W3cCredential,
            retries=retries,
        )

    async def _get_cred_def_objects(
        self, credential_definition_id: str
    ) -> tuple[Entry, Entry]:
        try:
            async with self.profile.session() as session:
                cred_def = await session.handle.fetch(
                    CATEGORY_CRED_DEF, credential_definition_id
                )
                cred_def_private = await session.handle.fetch(
                    CATEGORY_CRED_DEF_PRIVATE, credential_definition_id
                )
        except AskarError as err:
            raise AnonCredsRevocationError(
                "Error retrieving credential definition"
            ) from err
        if not cred_def or not cred_def_private:
            raise AnonCredsRevocationError(
                "Credential definition not found for credential issuance"
            )
        return cred_def, cred_def_private

    def _check_and_get_attribute_raw_values(
        self, schema_attributes: List[str], credential_values: dict
    ) -> Mapping[str, str]:
        raw_values = {}
        for attribute in schema_attributes:
            # Ensure every attribute present in schema to be set.
            # Extraneous attribute names are ignored.
            try:
                credential_value = credential_values[attribute]
            except KeyError:
                raise AnonCredsRevocationError(
                    "Provided credential values are missing a value "
                    f"for the schema attribute '{attribute}'"
                )

            raw_values[attribute] = str(credential_value)
        return raw_values

    async def _create_credential(
        self,
        credential_definition_id: str,
        schema_attributes: List[str],
        credential_offer: dict,
        credential_request: dict,
        credential_values: dict,
        credential_type: Union[Credential, W3cCredential],
        rev_reg_def_id: Optional[str] = None,
        tails_file_path: Optional[str] = None,
    ) -> Tuple[str, str]:
        """Create a credential.

        Args:
            credential_definition_id: The credential definition ID
            schema_attributes: The schema attributes
            credential_offer: The credential offer
            credential_request: The credential request
            credential_values: The credential values
            credential_type: The credential type
            rev_reg_def_id: The revocation registry definition ID
            tails_file_path: The tails file path

        Returns:
            A tuple of created credential and revocation ID

        """

        def _handle_missing_entries(rev_list: Entry, rev_reg_def: Entry, rev_key: Entry):
            if not rev_list:
                raise AnonCredsRevocationError("Revocation registry list not found")
            if not rev_reg_def:
                raise AnonCredsRevocationError("Revocation registry definition not found")
            if not rev_key:
                raise AnonCredsRevocationError(
                    "Revocation registry definition private data not found"
                )

        def _has_required_id_and_tails_path():
            return rev_reg_def_id and tails_file_path

        revoc = None
        credential_revocation_id = None
        rev_list = None

        if _has_required_id_and_tails_path():
            async with self.profile.session() as session:
                rev_reg_def = await session.handle.fetch(
                    CATEGORY_REV_REG_DEF, rev_reg_def_id
                )
                rev_list = await session.handle.fetch(CATEGORY_REV_LIST, rev_reg_def_id)
                rev_key = await session.handle.fetch(
                    CATEGORY_REV_REG_DEF_PRIVATE, rev_reg_def_id
                )

            _handle_missing_entries(rev_list, rev_reg_def, rev_key)

            rev_list_value_json = rev_list.value_json
            rev_list_tags = rev_list.tags

            # If the rev_list state is failed then the tails file was never uploaded,
            # try to upload it now and finish the revocation list
            if rev_list_tags.get("state") == RevListState.STATE_FAILED:
                await self.upload_tails_file(
                    RevRegDef.deserialize(rev_reg_def.value_json)
                )
                rev_list_tags["state"] = RevListState.STATE_FINISHED

            rev_reg_index = rev_list_value_json["next_index"]
            try:
                rev_reg_def = RevocationRegistryDefinition.load(rev_reg_def.raw_value)
                rev_list = RevocationStatusList.load(rev_list_value_json["rev_list"])
            except AnoncredsError as err:
                raise AnonCredsRevocationError(
                    "Error loading revocation registry"
                ) from err

            # NOTE: we increment the index ahead of time to keep the
            # transaction short. The revocation registry itself will NOT
            # be updated because we always use ISSUANCE_BY_DEFAULT.
            # If something goes wrong later, the index will be skipped.
            # FIXME - double check issuance type in case of upgraded wallet?
            if rev_reg_index > rev_reg_def.max_cred_num:
                raise AnonCredsRevocationRegistryFullError("Revocation registry is full")
            rev_list_value_json["next_index"] = rev_reg_index + 1
            async with self.profile.transaction() as txn:
                await txn.handle.replace(
                    CATEGORY_REV_LIST,
                    rev_reg_def_id,
                    value_json=rev_list_value_json,
                    tags=rev_list_tags,
                )
                await txn.commit()

            revoc = CredentialRevocationConfig(
                rev_reg_def,
                rev_key.raw_value,
                rev_list,
                rev_reg_index,
            )
            credential_revocation_id = str(rev_reg_index)

        cred_def, cred_def_private = await self._get_cred_def_objects(
            credential_definition_id
        )

        try:
            credential = await asyncio.get_event_loop().run_in_executor(
                None,
                lambda: credential_type.create(
                    cred_def=cred_def.raw_value,
                    cred_def_private=cred_def_private.raw_value,
                    cred_offer=credential_offer,
                    cred_request=credential_request,
                    attr_raw_values=self._check_and_get_attribute_raw_values(
                        schema_attributes, credential_values
                    ),
                    revocation_config=revoc,
                ),
            )
        except AnoncredsError as err:
            raise AnonCredsRevocationError("Error creating credential") from err

        return credential.to_json(), credential_revocation_id

    async def create_credential(
        self,
        credential_offer: dict,
        credential_request: dict,
        credential_values: dict,
        *,
        retries: int = 5,
    ) -> Tuple[str, str, str]:
        """Create a credential.

        Args:
            credential_offer: Credential Offer to create credential for
            credential_request: Credential request to create credential for
            credential_values: Values to go in credential
            revoc_reg_id: ID of the revocation registry
            retries: number of times to retry credential creation

        Returns:
            A tuple of created credential and revocation id

        """
        return await self._create_credential_helper(
            credential_offer,
            credential_request,
            credential_values,
            Credential,
            retries=retries,
        )

    async def _create_credential_helper(
        self,
        credential_offer: dict,
        credential_request: dict,
        credential_values: dict,
        credential_type: Union[Credential, W3cCredential],
        *,
        retries: int = 5,
    ) -> Tuple[str, str, str]:
        """Create a credential.

        Args:
            credential_offer: Credential Offer to create credential for
            credential_request: Credential request to create credential for
            credential_values: Values to go in credential
            credential_type: Credential or W3cCredential
            retries: number of times to retry credential creation

        Returns:
            A tuple of created credential and revocation id

        """
        issuer = AnonCredsIssuer(self.profile)
        anoncreds_registry = self.profile.inject(AnonCredsRegistry)
        schema_id = credential_offer["schema_id"]
        schema_result = await anoncreds_registry.get_schema(self.profile, schema_id)
        cred_def_id = credential_offer["cred_def_id"]

        revocable = await issuer.cred_def_supports_revocation(cred_def_id)

        for attempt in range(max(retries, 1)):
            if attempt > 0:
                LOGGER.info(
                    "Waiting 2s before retrying credential issuance for cred def '%s'",
                    cred_def_id,
                )
                await asyncio.sleep(2)

            rev_reg_def_result = None
            if revocable:
                rev_reg_def_result = await self.get_or_create_active_registry(cred_def_id)
                if (
                    rev_reg_def_result.revocation_registry_definition_state.state
                    != STATE_FINISHED
                ):
                    continue
                rev_reg_def_id = rev_reg_def_result.rev_reg_def_id
                tails_file_path = self.get_local_tails_path(
                    rev_reg_def_result.rev_reg_def
                )
            else:
                rev_reg_def_id = None
                tails_file_path = None

            try:
                cred_json, cred_rev_id = await self._create_credential(
                    cred_def_id,
                    schema_result.schema_value.attr_names,
                    credential_offer,
                    credential_request,
                    credential_values,
                    credential_type,
                    rev_reg_def_id,
                    tails_file_path,
                )
            except AnonCredsRevocationError as err:
                LOGGER.warning(f"Failed to create credential: {err.message}, retrying")
                continue

            def _is_full_registry(
                rev_reg_def_result: RevRegDefResult, cred_rev_id: str
            ) -> bool:
                # if we wait until max cred num is reached, we are too late.
                return (
                    rev_reg_def_result.rev_reg_def.value.max_cred_num
                    <= int(cred_rev_id) + 1
                )

            if rev_reg_def_result and _is_full_registry(rev_reg_def_result, cred_rev_id):
                await self.handle_full_registry(rev_reg_def_id)

            return cred_json, cred_rev_id, rev_reg_def_id

        raise AnonCredsRevocationError(
            f"Cred def '{cred_def_id}' revocation registry or list is in a bad state"
        )

    async def revoke_pending_credentials(
        self,
        revoc_reg_id: str,
        *,
        additional_crids: Optional[Sequence[int]] = None,
        limit_crids: Optional[Sequence[int]] = None,
    ) -> RevokeResult:
        """Revoke a set of credentials in a revocation registry.

        Args:
            revoc_reg_id: ID of the revocation registry
            additional_crids: sequences of additional credential indexes to revoke
            limit_crids: a sequence of credential indexes to limit revocation to
                If None, all pending revocations will be published.
                If given, the intersection of pending and limit crids will be published.

        Returns:
            Tuple with the update revocation list, list of cred rev ids not revoked

        """
        LOGGER.info(
            "Starting revocation process for registry %s with "
            "additional_crids=%s, limit_crids=%s",
            revoc_reg_id,
            additional_crids,
            limit_crids,
        )
        updated_list = None
        failed_crids = set()
        max_attempt = 5
        attempt = 0

        while True:
            attempt += 1
            LOGGER.debug("Revocation attempt %d/%d", attempt, max_attempt)
            if attempt >= max_attempt:
                LOGGER.error(
                    "Max attempts (%d) reached while trying to update registry %s",
                    max_attempt,
                    revoc_reg_id,
                )
                raise AnonCredsRevocationError(
                    "Repeated conflict attempting to update registry"
                )
            try:
                async with self.profile.session() as session:
                    LOGGER.debug("Fetching revocation registry data for %s", revoc_reg_id)
                    rev_reg_def_entry = await session.handle.fetch(
                        CATEGORY_REV_REG_DEF, revoc_reg_id
                    )
                    rev_list_entry = await session.handle.fetch(
                        CATEGORY_REV_LIST, revoc_reg_id
                    )
                    rev_reg_def_private_entry = await session.handle.fetch(
                        CATEGORY_REV_REG_DEF_PRIVATE, revoc_reg_id
                    )
            except AskarError as err:
                LOGGER.error(
                    "Failed to retrieve revocation registry data for %s: %s",
                    revoc_reg_id,
                    str(err),
                )
                raise AnonCredsRevocationError(
                    "Error retrieving revocation registry"
                ) from err

            if (
                not rev_reg_def_entry
                or not rev_list_entry
                or not rev_reg_def_private_entry
            ):
                missing_data = []
                if not rev_reg_def_entry:
                    missing_data.append("revocation registry definition")
                if not rev_list_entry:
                    missing_data.append("revocation list")
                if not rev_reg_def_private_entry:
                    missing_data.append("revocation registry private definition")
                LOGGER.error(
                    "Missing required revocation registry data for %s: %s",
                    revoc_reg_id,
                    ", ".join(missing_data),
                )
                raise AnonCredsRevocationError(
                    f"Missing required revocation registry data: {' '.join(missing_data)}"
                )

            try:
                async with self.profile.session() as session:
                    cred_def_id = rev_reg_def_entry.value_json["credDefId"]
                    LOGGER.debug("Fetching credential definition %s", cred_def_id)
                    cred_def_entry = await session.handle.fetch(
                        CATEGORY_CRED_DEF, cred_def_id
                    )
            except AskarError as err:
                LOGGER.error(
                    "Failed to retrieve credential definition %s: %s",
                    cred_def_id,
                    str(err),
                )
                raise AnonCredsRevocationError(
                    f"Error retrieving cred def {cred_def_id}"
                ) from err

            try:
                # TODO This is a little rough; stored tails location will have public uri
                # but library needs local tails location
                LOGGER.debug("Deserializing revocation registry data")
                rev_reg_def = RevRegDef.deserialize(rev_reg_def_entry.value_json)
                rev_reg_def.value.tails_location = self.get_local_tails_path(rev_reg_def)
                cred_def = CredDef.deserialize(cred_def_entry.value_json)
                rev_reg_def_private = RevocationRegistryDefinitionPrivate.load(
                    rev_reg_def_private_entry.value_json
                )
            except AnoncredsError as err:
                LOGGER.error(
                    "Failed to load revocation registry definition: %s", str(err)
                )
                raise AnonCredsRevocationError(
                    "Error loading revocation registry definition"
                ) from err

            rev_crids = set()
            failed_crids = set()
            max_cred_num = rev_reg_def.value.max_cred_num
            rev_info = rev_list_entry.value_json
            cred_revoc_ids = (rev_info["pending"] or []) + (additional_crids or [])
            rev_list = RevList.deserialize(rev_info["rev_list"])

            LOGGER.info(
                "Processing %d credential revocation IDs for registry %s",
                len(cred_revoc_ids),
                revoc_reg_id,
            )

            for rev_id in cred_revoc_ids:
                if rev_id < 1 or rev_id > max_cred_num:
                    LOGGER.error(
                        "Skipping requested credential revocation "
                        "on rev reg id %s, cred rev id=%s not in range (1-%d)",
                        revoc_reg_id,
                        rev_id,
                        max_cred_num,
                    )
                    failed_crids.add(rev_id)
                elif rev_id >= rev_info["next_index"]:
                    LOGGER.warning(
                        "Skipping requested credential revocation "
                        "on rev reg id %s, cred rev id=%s not yet issued (next_index=%d)",
                        revoc_reg_id,
                        rev_id,
                        rev_info["next_index"],
                    )
                    failed_crids.add(rev_id)
                elif rev_list.revocation_list[rev_id] == 1:
                    LOGGER.warning(
                        "Skipping requested credential revocation "
                        "on rev reg id %s, cred rev id=%s already revoked",
                        revoc_reg_id,
                        rev_id,
                    )
                    failed_crids.add(rev_id)
                else:
                    rev_crids.add(rev_id)

            if not rev_crids:
                LOGGER.info(
                    "No valid credentials to revoke for registry %s", revoc_reg_id
                )
                break

            if limit_crids is None or limit_crids == []:
                skipped_crids = set()
            else:
                skipped_crids = rev_crids - set(limit_crids)
            rev_crids = rev_crids - skipped_crids

            LOGGER.info(
                "Revoking %d credentials, skipping %d credentials for registry %s",
                len(rev_crids),
                len(skipped_crids),
                revoc_reg_id,
            )

            try:
                LOGGER.debug("Updating revocation list with new revocations")
                updated_list = await asyncio.get_event_loop().run_in_executor(
                    None,
                    lambda: rev_list.to_native().update(
                        cred_def=cred_def.to_native(),
                        rev_reg_def=rev_reg_def.to_native(),
                        rev_reg_def_private=rev_reg_def_private,
                        issued=None,
                        revoked=list(rev_crids),
                        timestamp=int(time.time()),
                    ),
                )
            except AnoncredsError as err:
                LOGGER.error("Failed to update revocation registry: %s", str(err))
                raise AnonCredsRevocationError(
                    "Error updating revocation registry"
                ) from err

            try:
                async with self.profile.transaction() as txn:
                    LOGGER.debug("Saving updated revocation list")
                    rev_info_upd = await txn.handle.fetch(
                        CATEGORY_REV_LIST, revoc_reg_id, for_update=True
                    )
                    if not rev_info_upd:
                        LOGGER.warning(
                            "Revocation registry %s missing during update, skipping",
                            revoc_reg_id,
                        )
                        updated_list = None
                        break
                    tags = rev_info_upd.tags
                    rev_info_upd = rev_info_upd.value_json
                    if rev_info_upd != rev_info:
                        LOGGER.debug(
                            "Concurrent update detected for registry %s, retrying",
                            revoc_reg_id,
                        )
                        continue
                    rev_info_upd["rev_list"] = updated_list.to_dict()
                    rev_info_upd["pending"] = (
                        list(skipped_crids) if skipped_crids else None
                    )
                    tags["pending"] = json.dumps(True if skipped_crids else False)
                    await txn.handle.replace(
                        CATEGORY_REV_LIST,
                        revoc_reg_id,
                        value_json=rev_info_upd,
                        tags=tags,
                    )
                    await txn.commit()
                    LOGGER.info(
                        "Successfully updated revocation list for registry %s",
                        revoc_reg_id,
                    )
            except AskarError as err:
                LOGGER.error("Failed to save revocation registry: %s", str(err))
                raise AnonCredsRevocationError(
                    "Error saving revocation registry"
                ) from err
            break

        result = RevokeResult(
            prev=rev_list,
            curr=RevList.from_native(updated_list) if updated_list else None,
            revoked=list(rev_crids),
            failed=[str(rev_id) for rev_id in sorted(failed_crids)],
        )
        LOGGER.info(
            "Completed revocation process for registry %s: %d revoked, %d failed",
            revoc_reg_id,
            len(result.revoked),
            len(result.failed),
        )
        return result

    async def mark_pending_revocations(self, rev_reg_def_id: str, *crids: int) -> None:
        """Cred rev ids stored to publish later."""
        async with self.profile.transaction() as txn:
            entry = await txn.handle.fetch(
                CATEGORY_REV_LIST,
                rev_reg_def_id,
                for_update=True,
            )

            if not entry:
                raise AnonCredsRevocationError(
                    "Revocation list with id {rev_reg_def_id} not found"
                )

            pending: Optional[List[int]] = entry.value_json["pending"]
            if pending:
                pending.extend(crids)
            else:
                pending = list(crids)

            value = entry.value_json
            value["pending"] = pending
            tags = entry.tags
            tags["pending"] = json.dumps(True)
            await txn.handle.replace(
                CATEGORY_REV_LIST,
                rev_reg_def_id,
                value_json=value,
                tags=tags,
            )
            await txn.commit()

    async def get_pending_revocations(self, rev_reg_def_id: str) -> List[int]:
        """Retrieve the list of credential revocation ids pending revocation."""
        async with self.profile.session() as session:
            entry = await session.handle.fetch(CATEGORY_REV_LIST, rev_reg_def_id)
            if not entry:
                return []

            return entry.value_json["pending"] or []

    async def clear_pending_revocations(
        self,
        txn: ProfileSession,
        rev_reg_def_id: str,
        crid_mask: Optional[Sequence[int]] = None,
    ) -> None:
        """Clear pending revocations."""
        if not isinstance(txn, AskarAnonCredsProfileSession):
            raise ValueError("Askar wallet required")

        entry = await txn.handle.fetch(
            CATEGORY_REV_LIST,
            rev_reg_def_id,
            for_update=True,
        )

        if not entry:
            raise AnonCredsRevocationError(
                "Revocation list with id {rev_reg_def_id} not found"
            )

        value = entry.value_json
        if crid_mask is None:
            value["pending"] = None
        else:
            value["pending"] = set(value["pending"]) - set(crid_mask)

        tags = entry.tags
        tags["pending"] = json.dumps(False)
        await txn.handle.replace(
            CATEGORY_REV_LIST,
            rev_reg_def_id,
            value_json=value,
            tags=tags,
        )

    async def set_tails_file_public_uri(self, rev_reg_id: str, tails_public_uri: str):
        """Update Revocation Registry tails file public uri."""
        # TODO: Implement or remove
        pass

    async def set_rev_reg_state(self, rev_reg_id: str, state: str):
        """Update Revocation Registry state."""
        # TODO: Implement or remove
        pass



"""Message type identifiers for Trust Pings."""

from ...didcomm_prefix import DIDCommPrefix

SPEC_URI = "https://github.com/decentralized-identity/aries-rfcs/tree/main/features/0794-did-rotate"

# Message types
ROTATE = "did-rotate/1.0/rotate"
ACK = "did-rotate/1.0/ack"
HANGUP = "did-rotate/1.0/hangup"
PROBLEM_REPORT = "did-rotate/1.0/problem-report"

PROTOCOL_PACKAGE = "acapy_agent.protocols.did_rotate.v1_0"

MESSAGE_TYPES = DIDCommPrefix.qualify_all(
    {
        ROTATE: f"{PROTOCOL_PACKAGE}.messages.rotate.Rotate",
        ACK: f"{PROTOCOL_PACKAGE}.messages.ack.RotateAck",
        HANGUP: f"{PROTOCOL_PACKAGE}.messages.hangup.Hangup",
        PROBLEM_REPORT: f"{PROTOCOL_PACKAGE}.messages.problem_report.RotateProblemReport",
    }
)



SECURITY_V3_UNSTABLE = {
    "@context": [
        {
            "@version": 1.1,
            "id": "@id",
            "type": "@type",
            "@protected": True,
            "JsonWebKey2020": {"@id": "https://w3id.org/security#JsonWebKey2020"},
            "JsonWebSignature2020": {
                "@id": "https://w3id.org/security#JsonWebSignature2020",
                "@context": {
                    "@version": 1.1,
                    "id": "@id",
                    "type": "@type",
                    "@protected": True,
                    "challenge": "https://w3id.org/security#challenge",
                    "created": {
                        "@id": "http://purl.org/dc/terms/created",
                        "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                    },
                    "domain": "https://w3id.org/security#domain",
                    "expires": {
                        "@id": "https://w3id.org/security#expiration",
                        "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                    },
                    "jws": "https://w3id.org/security#jws",
                    "nonce": "https://w3id.org/security#nonce",
                    "proofPurpose": {
                        "@id": "https://w3id.org/security#proofPurpose",
                        "@type": "@vocab",
                        "@context": {
                            "@version": 1.1,
                            "@protected": True,
                            "id": "@id",
                            "type": "@type",
                            "assertionMethod": {
                                "@id": "https://w3id.org/security#assertionMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "authentication": {
                                "@id": "https://w3id.org/security#authenticationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "capabilityInvocation": {
                                "@id": "https://w3id.org/security#capabilityInvocationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "capabilityDelegation": {
                                "@id": "https://w3id.org/security#capabilityDelegationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "keyAgreement": {
                                "@id": "https://w3id.org/security#keyAgreementMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                        },
                    },
                    "verificationMethod": {
                        "@id": "https://w3id.org/security#verificationMethod",
                        "@type": "@id",
                    },
                },
            },
            "Ed25519VerificationKey2020": {
                "@id": "https://w3id.org/security#Ed25519VerificationKey2020"
            },
            "Ed25519Signature2020": {
                "@id": "https://w3id.org/security#Ed25519Signature2020",
                "@context": {
                    "@protected": True,
                    "id": "@id",
                    "type": "@type",
                    "challenge": "https://w3id.org/security#challenge",
                    "created": {
                        "@id": "http://purl.org/dc/terms/created",
                        "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                    },
                    "domain": "https://w3id.org/security#domain",
                    "expires": {
                        "@id": "https://w3id.org/security#expiration",
                        "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                    },
                    "nonce": "https://w3id.org/security#nonce",
                    "proofPurpose": {
                        "@id": "https://w3id.org/security#proofPurpose",
                        "@type": "@vocab",
                        "@context": {
                            "@version": 1.1,
                            "@protected": True,
                            "id": "@id",
                            "type": "@type",
                            "assertionMethod": {
                                "@id": "https://w3id.org/security#assertionMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "authentication": {
                                "@id": "https://w3id.org/security#authenticationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "capabilityInvocation": {
                                "@id": "https://w3id.org/security#capabilityInvocationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "capabilityDelegation": {
                                "@id": "https://w3id.org/security#capabilityDelegationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "keyAgreement": {
                                "@id": "https://w3id.org/security#keyAgreementMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                        },
                    },
                    "proofValue": {
                        "@id": "https://w3id.org/security#proofValue",
                        "@type": "https://w3id.org/security#multibase",
                    },
                    "verificationMethod": {
                        "@id": "https://w3id.org/security#verificationMethod",
                        "@type": "@id",
                    },
                },
            },
            "publicKeyJwk": {
                "@id": "https://w3id.org/security#publicKeyJwk",
                "@type": "@json",
            },
            "ethereumAddress": {"@id": "https://w3id.org/security#ethereumAddress"},
            "publicKeyHex": {"@id": "https://w3id.org/security#publicKeyHex"},
            "blockchainAccountId": {
                "@id": "https://w3id.org/security#blockchainAccountId"
            },
            "MerkleProof2019": {"@id": "https://w3id.org/security#MerkleProof2019"},
            "Bls12381G1Key2020": {"@id": "https://w3id.org/security#Bls12381G1Key2020"},
            "Bls12381G2Key2020": {"@id": "https://w3id.org/security#Bls12381G2Key2020"},
            "BbsBlsSignature2020": {
                "@id": "https://w3id.org/security#BbsBlsSignature2020",
                "@context": {
                    "@protected": True,
                    "id": "@id",
                    "type": "@type",
                    "challenge": "https://w3id.org/security#challenge",
                    "created": {
                        "@id": "http://purl.org/dc/terms/created",
                        "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                    },
                    "domain": "https://w3id.org/security#domain",
                    "nonce": "https://w3id.org/security#nonce",
                    "proofPurpose": {
                        "@id": "https://w3id.org/security#proofPurpose",
                        "@type": "@vocab",
                        "@context": {
                            "@version": 1.1,
                            "@protected": True,
                            "id": "@id",
                            "type": "@type",
                            "assertionMethod": {
                                "@id": "https://w3id.org/security#assertionMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "authentication": {
                                "@id": "https://w3id.org/security#authenticationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "capabilityInvocation": {
                                "@id": "https://w3id.org/security#capabilityInvocationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "capabilityDelegation": {
                                "@id": "https://w3id.org/security#capabilityDelegationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "keyAgreement": {
                                "@id": "https://w3id.org/security#keyAgreementMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                        },
                    },
                    "proofValue": "https://w3id.org/security#proofValue",
                    "verificationMethod": {
                        "@id": "https://w3id.org/security#verificationMethod",
                        "@type": "@id",
                    },
                },
            },
            "BbsBlsSignatureProof2020": {
                "@id": "https://w3id.org/security#BbsBlsSignatureProof2020",
                "@context": {
                    "@protected": True,
                    "id": "@id",
                    "type": "@type",
                    "challenge": "https://w3id.org/security#challenge",
                    "created": {
                        "@id": "http://purl.org/dc/terms/created",
                        "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                    },
                    "domain": "https://w3id.org/security#domain",
                    "nonce": "https://w3id.org/security#nonce",
                    "proofPurpose": {
                        "@id": "https://w3id.org/security#proofPurpose",
                        "@type": "@vocab",
                        "@context": {
                            "@version": 1.1,
                            "@protected": True,
                            "id": "@id",
                            "type": "@type",
                            "assertionMethod": {
                                "@id": "https://w3id.org/security#assertionMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "authentication": {
                                "@id": "https://w3id.org/security#authenticationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "capabilityInvocation": {
                                "@id": "https://w3id.org/security#capabilityInvocationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "capabilityDelegation": {
                                "@id": "https://w3id.org/security#capabilityDelegationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "keyAgreement": {
                                "@id": "https://w3id.org/security#keyAgreementMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                        },
                    },
                    "proofValue": "https://w3id.org/security#proofValue",
                    "verificationMethod": {
                        "@id": "https://w3id.org/security#verificationMethod",
                        "@type": "@id",
                    },
                },
            },
            "EcdsaKoblitzSignature2016": "https://w3id.org/security#EcdsaKoblitzSignature2016",
            "Ed25519Signature2018": {
                "@id": "https://w3id.org/security#Ed25519Signature2018",
                "@context": {
                    "@protected": True,
                    "id": "@id",
                    "type": "@type",
                    "challenge": "https://w3id.org/security#challenge",
                    "created": {
                        "@id": "http://purl.org/dc/terms/created",
                        "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                    },
                    "domain": "https://w3id.org/security#domain",
                    "expires": {
                        "@id": "https://w3id.org/security#expiration",
                        "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                    },
                    "jws": "https://w3id.org/security#jws",
                    "nonce": "https://w3id.org/security#nonce",
                    "proofPurpose": {
                        "@id": "https://w3id.org/security#proofPurpose",
                        "@type": "@vocab",
                        "@context": {
                            "@version": 1.1,
                            "@protected": True,
                            "id": "@id",
                            "type": "@type",
                            "assertionMethod": {
                                "@id": "https://w3id.org/security#assertionMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "authentication": {
                                "@id": "https://w3id.org/security#authenticationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "capabilityInvocation": {
                                "@id": "https://w3id.org/security#capabilityInvocationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "capabilityDelegation": {
                                "@id": "https://w3id.org/security#capabilityDelegationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "keyAgreement": {
                                "@id": "https://w3id.org/security#keyAgreementMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                        },
                    },
                    "proofValue": "https://w3id.org/security#proofValue",
                    "verificationMethod": {
                        "@id": "https://w3id.org/security#verificationMethod",
                        "@type": "@id",
                    },
                },
            },
            "EncryptedMessage": "https://w3id.org/security#EncryptedMessage",
            "GraphSignature2012": "https://w3id.org/security#GraphSignature2012",
            "LinkedDataSignature2015": "https://w3id.org/security#LinkedDataSignature2015",
            "LinkedDataSignature2016": "https://w3id.org/security#LinkedDataSignature2016",
            "CryptographicKey": "https://w3id.org/security#Key",
            "authenticationTag": "https://w3id.org/security#authenticationTag",
            "canonicalizationAlgorithm": "https://w3id.org/security#canonicalizationAlgorithm",
            "cipherAlgorithm": "https://w3id.org/security#cipherAlgorithm",
            "cipherData": "https://w3id.org/security#cipherData",
            "cipherKey": "https://w3id.org/security#cipherKey",
            "created": {
                "@id": "http://purl.org/dc/terms/created",
                "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
            },
            "creator": {"@id": "http://purl.org/dc/terms/creator", "@type": "@id"},
            "digestAlgorithm": "https://w3id.org/security#digestAlgorithm",
            "digestValue": "https://w3id.org/security#digestValue",
            "domain": "https://w3id.org/security#domain",
            "encryptionKey": "https://w3id.org/security#encryptionKey",
            "expiration": {
                "@id": "https://w3id.org/security#expiration",
                "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
            },
            "expires": {
                "@id": "https://w3id.org/security#expiration",
                "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
            },
            "initializationVector": "https://w3id.org/security#initializationVector",
            "iterationCount": "https://w3id.org/security#iterationCount",
            "nonce": "https://w3id.org/security#nonce",
            "normalizationAlgorithm": "https://w3id.org/security#normalizationAlgorithm",
            "owner": "https://w3id.org/security#owner",
            "password": "https://w3id.org/security#password",
            "privateKey": "https://w3id.org/security#privateKey",
            "privateKeyPem": "https://w3id.org/security#privateKeyPem",
            "publicKey": "https://w3id.org/security#publicKey",
            "publicKeyBase58": "https://w3id.org/security#publicKeyBase58",
            "publicKeyPem": "https://w3id.org/security#publicKeyPem",
            "publicKeyWif": "https://w3id.org/security#publicKeyWif",
            "publicKeyService": "https://w3id.org/security#publicKeyService",
            "revoked": {
                "@id": "https://w3id.org/security#revoked",
                "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
            },
            "salt": "https://w3id.org/security#salt",
            "signature": "https://w3id.org/security#signature",
            "signatureAlgorithm": "https://w3id.org/security#signingAlgorithm",
            "signatureValue": "https://w3id.org/security#signatureValue",
            "proofValue": "https://w3id.org/security#proofValue",
            "AesKeyWrappingKey2019": "https://w3id.org/security#AesKeyWrappingKey2019",
            "DeleteKeyOperation": "https://w3id.org/security#DeleteKeyOperation",
            "DeriveSecretOperation": "https://w3id.org/security#DeriveSecretOperation",
            "EcdsaSecp256k1Signature2019": {
                "@id": "https://w3id.org/security#EcdsaSecp256k1Signature2019",
                "@context": {
                    "@protected": True,
                    "id": "@id",
                    "type": "@type",
                    "challenge": "https://w3id.org/security#challenge",
                    "created": {
                        "@id": "http://purl.org/dc/terms/created",
                        "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                    },
                    "domain": "https://w3id.org/security#domain",
                    "expires": {
                        "@id": "https://w3id.org/security#expiration",
                        "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                    },
                    "jws": "https://w3id.org/security#jws",
                    "nonce": "https://w3id.org/security#nonce",
                    "proofPurpose": {
                        "@id": "https://w3id.org/security#proofPurpose",
                        "@type": "@vocab",
                        "@context": {
                            "@version": 1.1,
                            "@protected": True,
                            "id": "@id",
                            "type": "@type",
                            "assertionMethod": {
                                "@id": "https://w3id.org/security#assertionMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "authentication": {
                                "@id": "https://w3id.org/security#authenticationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "capabilityInvocation": {
                                "@id": "https://w3id.org/security#capabilityInvocationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "capabilityDelegation": {
                                "@id": "https://w3id.org/security#capabilityDelegationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "keyAgreement": {
                                "@id": "https://w3id.org/security#keyAgreementMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                        },
                    },
                    "proofValue": "https://w3id.org/security#proofValue",
                    "verificationMethod": {
                        "@id": "https://w3id.org/security#verificationMethod",
                        "@type": "@id",
                    },
                },
            },
            "EcdsaSecp256r1Signature2019": {
                "@id": "https://w3id.org/security#EcdsaSecp256r1Signature2019",
                "@context": {
                    "@protected": True,
                    "id": "@id",
                    "type": "@type",
                    "challenge": "https://w3id.org/security#challenge",
                    "created": {
                        "@id": "http://purl.org/dc/terms/created",
                        "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                    },
                    "domain": "https://w3id.org/security#domain",
                    "expires": {
                        "@id": "https://w3id.org/security#expiration",
                        "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                    },
                    "jws": "https://w3id.org/security#jws",
                    "nonce": "https://w3id.org/security#nonce",
                    "proofPurpose": {
                        "@id": "https://w3id.org/security#proofPurpose",
                        "@type": "@vocab",
                        "@context": {
                            "@version": 1.1,
                            "@protected": True,
                            "id": "@id",
                            "type": "@type",
                            "assertionMethod": {
                                "@id": "https://w3id.org/security#assertionMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "authentication": {
                                "@id": "https://w3id.org/security#authenticationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "capabilityInvocation": {
                                "@id": "https://w3id.org/security#capabilityInvocationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "capabilityDelegation": {
                                "@id": "https://w3id.org/security#capabilityDelegationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "keyAgreement": {
                                "@id": "https://w3id.org/security#keyAgreementMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                        },
                    },
                    "proofValue": "https://w3id.org/security#proofValue",
                    "verificationMethod": {
                        "@id": "https://w3id.org/security#verificationMethod",
                        "@type": "@id",
                    },
                },
            },
            "EcdsaSecp256k1VerificationKey2019": "https://w3id.org/security#EcdsaSecp256k1VerificationKey2019",
            "EcdsaSecp256r1VerificationKey2019": "https://w3id.org/security#EcdsaSecp256r1VerificationKey2019",
            "Ed25519VerificationKey2018": "https://w3id.org/security#Ed25519VerificationKey2018",
            "EquihashProof2018": "https://w3id.org/security#EquihashProof2018",
            "ExportKeyOperation": "https://w3id.org/security#ExportKeyOperation",
            "GenerateKeyOperation": "https://w3id.org/security#GenerateKeyOperation",
            "KmsOperation": "https://w3id.org/security#KmsOperation",
            "RevokeKeyOperation": "https://w3id.org/security#RevokeKeyOperation",
            "RsaSignature2018": {
                "@id": "https://w3id.org/security#RsaSignature2018",
                "@context": {
                    "@protected": True,
                    "challenge": "https://w3id.org/security#challenge",
                    "created": {
                        "@id": "http://purl.org/dc/terms/created",
                        "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                    },
                    "domain": "https://w3id.org/security#domain",
                    "expires": {
                        "@id": "https://w3id.org/security#expiration",
                        "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                    },
                    "jws": "https://w3id.org/security#jws",
                    "nonce": "https://w3id.org/security#nonce",
                    "proofPurpose": {
                        "@id": "https://w3id.org/security#proofPurpose",
                        "@type": "@vocab",
                        "@context": {
                            "@version": 1.1,
                            "@protected": True,
                            "id": "@id",
                            "type": "@type",
                            "assertionMethod": {
                                "@id": "https://w3id.org/security#assertionMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "authentication": {
                                "@id": "https://w3id.org/security#authenticationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "capabilityInvocation": {
                                "@id": "https://w3id.org/security#capabilityInvocationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "capabilityDelegation": {
                                "@id": "https://w3id.org/security#capabilityDelegationMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                            "keyAgreement": {
                                "@id": "https://w3id.org/security#keyAgreementMethod",
                                "@type": "@id",
                                "@container": "@set",
                            },
                        },
                    },
                    "proofValue": "https://w3id.org/security#proofValue",
                    "verificationMethod": {
                        "@id": "https://w3id.org/security#verificationMethod",
                        "@type": "@id",
                    },
                },
            },
            "RsaVerificationKey2018": "https://w3id.org/security#RsaVerificationKey2018",
            "Sha256HmacKey2019": "https://w3id.org/security#Sha256HmacKey2019",
            "SignOperation": "https://w3id.org/security#SignOperation",
            "UnwrapKeyOperation": "https://w3id.org/security#UnwrapKeyOperation",
            "VerifyOperation": "https://w3id.org/security#VerifyOperation",
            "WrapKeyOperation": "https://w3id.org/security#WrapKeyOperation",
            "X25519KeyAgreementKey2019": "https://w3id.org/security#X25519KeyAgreementKey2019",
            "allowedAction": "https://w3id.org/security#allowedAction",
            "assertionMethod": {
                "@id": "https://w3id.org/security#assertionMethod",
                "@type": "@id",
                "@container": "@set",
            },
            "authentication": {
                "@id": "https://w3id.org/security#authenticationMethod",
                "@type": "@id",
                "@container": "@set",
            },
            "capability": {
                "@id": "https://w3id.org/security#capability",
                "@type": "@id",
            },
            "capabilityAction": "https://w3id.org/security#capabilityAction",
            "capabilityChain": {
                "@id": "https://w3id.org/security#capabilityChain",
                "@type": "@id",
                "@container": "@list",
            },
            "capabilityDelegation": {
                "@id": "https://w3id.org/security#capabilityDelegationMethod",
                "@type": "@id",
                "@container": "@set",
            },
            "capabilityInvocation": {
                "@id": "https://w3id.org/security#capabilityInvocationMethod",
                "@type": "@id",
                "@container": "@set",
            },
            "caveat": {
                "@id": "https://w3id.org/security#caveat",
                "@type": "@id",
                "@container": "@set",
            },
            "challenge": "https://w3id.org/security#challenge",
            "ciphertext": "https://w3id.org/security#ciphertext",
            "controller": {
                "@id": "https://w3id.org/security#controller",
                "@type": "@id",
            },
            "delegator": {"@id": "https://w3id.org/security#delegator", "@type": "@id"},
            "equihashParameterK": {
                "@id": "https://w3id.org/security#equihashParameterK",
                "@type": "http://www.w3.org/2001/XMLSchema#:integer",
            },
            "equihashParameterN": {
                "@id": "https://w3id.org/security#equihashParameterN",
                "@type": "http://www.w3.org/2001/XMLSchema#:integer",
            },
            "invocationTarget": {
                "@id": "https://w3id.org/security#invocationTarget",
                "@type": "@id",
            },
            "invoker": {"@id": "https://w3id.org/security#invoker", "@type": "@id"},
            "jws": "https://w3id.org/security#jws",
            "keyAgreement": {
                "@id": "https://w3id.org/security#keyAgreementMethod",
                "@type": "@id",
                "@container": "@set",
            },
            "kmsModule": {"@id": "https://w3id.org/security#kmsModule"},
            "parentCapability": {
                "@id": "https://w3id.org/security#parentCapability",
                "@type": "@id",
            },
            "plaintext": "https://w3id.org/security#plaintext",
            "proof": {
                "@id": "https://w3id.org/security#proof",
                "@type": "@id",
                "@container": "@graph",
            },
            "proofPurpose": {
                "@id": "https://w3id.org/security#proofPurpose",
                "@type": "@vocab",
                "@context": {
                    "@version": 1.1,
                    "@protected": True,
                    "id": "@id",
                    "type": "@type",
                    "assertionMethod": {
                        "@id": "https://w3id.org/security#assertionMethod",
                        "@type": "@id",
                        "@container": "@set",
                    },
                    "authentication": {
                        "@id": "https://w3id.org/security#authenticationMethod",
                        "@type": "@id",
                        "@container": "@set",
                    },
                    "capabilityInvocation": {
                        "@id": "https://w3id.org/security#capabilityInvocationMethod",
                        "@type": "@id",
                        "@container": "@set",
                    },
                    "capabilityDelegation": {
                        "@id": "https://w3id.org/security#capabilityDelegationMethod",
                        "@type": "@id",
                        "@container": "@set",
                    },
                    "keyAgreement": {
                        "@id": "https://w3id.org/security#keyAgreementMethod",
                        "@type": "@id",
                        "@container": "@set",
                    },
                },
            },
            "referenceId": "https://w3id.org/security#referenceId",
            "unwrappedKey": "https://w3id.org/security#unwrappedKey",
            "verificationMethod": {
                "@id": "https://w3id.org/security#verificationMethod",
                "@type": "@id",
            },
            "verifyData": "https://w3id.org/security#verifyData",
            "wrappedKey": "https://w3id.org/security#wrappedKey",
        }
    ]
}



import pytest
import pytest_asyncio

from .....connections.models.conn_record import ConnRecord
from .....core.profile import Profile
from .....messaging.responder import BaseResponder, MockResponder
from .....storage.error import StorageNotFoundError
from .....tests import mock
from .....utils.testing import create_test_profile
from .....wallet.askar import AskarWallet
from .....wallet.did_info import DIDInfo
from .....wallet.did_method import SOV
from .....wallet.key_type import ED25519
from ....routing.v1_0.models.route_record import RouteRecord
from ..manager import MediationManager
from ..messages.keylist_update import KeylistUpdate
from ..models.mediation_record import MediationRecord
from ..route_manager import (
    CoordinateMediationV1RouteManager,
    RouteManager,
    RouteManagerError,
)

TEST_RECORD_DID = "55GkHamhTU1ZbTbV2ab9DE"
TEST_RECORD_VERKEY = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
TEST_VERKEY = "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
TEST_ROUTE_RECORD_VERKEY = "9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"
TEST_ROUTE_VERKEY = "did:key:z6MknxTj6Zj1VrDWc1ofaZtmCVv2zNXpD58Xup4ijDGoQhya"
TEST_ROUTE_VERKEY = "did:key:z6MknxTj6Zj1VrDWc1ofaZtmCVv2zNXpD58Xup4ijDGoQhya"
TEST_ROUTE_VERKEY_REF = "did:key:z6MknxTj6Zj1VrDWc1ofaZtmCVv2zNXpD58Xup4ijDGoQhya#z6MknxTj6Zj1VrDWc1ofaZtmCVv2zNXpD58Xup4ijDGoQhya"
TEST_ROUTE_VERKEY_REF2 = "did:key:z6MknxTj6Zj1VrDWc1ofaZtmCVv2zNXpD58Xup4ijDGoQhyz#z6MknxTj6Zj1VrDWc1ofaZtmCVv2zNXpD58Xup4ijDGoQhyz"


class MockRouteManager(RouteManager):
    """Concretion of RouteManager for testing."""

    _route_for_key = mock.CoroutineMock()
    routing_info = mock.CoroutineMock()


@pytest.fixture
def mock_responder():
    yield MockResponder()


@pytest_asyncio.fixture
async def profile(mock_responder: MockResponder):
    profile = await create_test_profile()
    profile.context.injector.bind_instance(BaseResponder, mock_responder)
    yield profile


@pytest.fixture
def route_manager():
    manager = MockRouteManager()
    manager._route_for_key = mock.CoroutineMock(
        return_value=mock.MagicMock(KeylistUpdate)
    )
    manager.routing_info = mock.CoroutineMock(return_value=([], None))
    yield manager


@pytest.fixture
def mediation_route_manager():
    yield CoordinateMediationV1RouteManager()


@pytest.fixture
def conn_record():
    record = ConnRecord(connection_id="12345")
    record.metadata_get = mock.CoroutineMock(return_value={})
    record.metadata_set = mock.CoroutineMock()
    yield record


@pytest.mark.asyncio
async def test_get_or_create_my_did_no_did(
    profile: Profile, route_manager: RouteManager, conn_record: ConnRecord
):
    conn_record.my_did = None
    mock_did_info = mock.MagicMock()
    with (
        mock.patch.object(
            AskarWallet,
            "create_local_did",
            mock.CoroutineMock(return_value=mock_did_info),
        ) as mock_create_local_did,
        mock.patch.object(conn_record, "save", mock.CoroutineMock()) as mock_save,
    ):
        info = await route_manager.get_or_create_my_did(profile, conn_record)
        assert mock_did_info == info
        mock_create_local_did.assert_called_once()
        mock_save.assert_called_once()


@pytest.mark.asyncio
async def test_get_or_create_my_did_existing_did(
    profile: Profile, route_manager: RouteManager, conn_record: ConnRecord
):
    conn_record.my_did = "test-did"
    mock_did_info = mock.MagicMock(DIDInfo)
    with mock.patch.object(
        AskarWallet, "get_local_did", mock.CoroutineMock(return_value=mock_did_info)
    ) as mock_get_local_did:
        info = await route_manager.get_or_create_my_did(profile, conn_record)
        assert mock_did_info == info
        mock_get_local_did.assert_called_once()


@pytest.mark.asyncio
async def test_mediation_record_for_connection_mediation_id(
    profile: Profile, route_manager: RouteManager, conn_record: ConnRecord
):
    mediation_record = MediationRecord(mediation_id="test-mediation-id")
    with (
        mock.patch.object(
            route_manager,
            "mediation_record_if_id",
            mock.CoroutineMock(return_value=mediation_record),
        ) as mock_mediation_record_if_id,
        mock.patch.object(
            route_manager, "save_mediator_for_connection", mock.CoroutineMock()
        ),
    ):
        assert await route_manager.mediation_records_for_connection(
            profile, conn_record, mediation_record.mediation_id
        ) == [mediation_record]
        mock_mediation_record_if_id.assert_called_once_with(
            profile, mediation_record.mediation_id, False
        )


@pytest.mark.asyncio
async def test_mediation_record_for_connection_mediation_metadata(
    profile: Profile, route_manager: RouteManager, conn_record: ConnRecord
):
    mediation_record = MediationRecord(mediation_id="test-mediation-id")
    conn_record.metadata_get.return_value = {
        MediationManager.METADATA_ID: mediation_record.mediation_id
    }
    with (
        mock.patch.object(
            route_manager,
            "mediation_record_if_id",
            mock.CoroutineMock(return_value=mediation_record),
        ) as mock_mediation_record_if_id,
        mock.patch.object(
            route_manager, "save_mediator_for_connection", mock.CoroutineMock()
        ),
    ):
        assert await route_manager.mediation_records_for_connection(
            profile, conn_record, "another-mediation-id"
        ) == [mediation_record]
        mock_mediation_record_if_id.assert_called_once_with(
            profile, mediation_record.mediation_id, False
        )


@pytest.mark.asyncio
async def test_mediation_record_for_connection_default(
    profile: Profile, route_manager: RouteManager, conn_record: ConnRecord
):
    mediation_record = MediationRecord(mediation_id="test-mediation-id")
    with (
        mock.patch.object(
            route_manager,
            "mediation_record_if_id",
            mock.CoroutineMock(return_value=mediation_record),
        ) as mock_mediation_record_if_id,
        mock.patch.object(
            route_manager, "save_mediator_for_connection", mock.CoroutineMock()
        ),
    ):
        assert await route_manager.mediation_records_for_connection(
            profile, conn_record, None, or_default=True
        ) == [mediation_record]
        mock_mediation_record_if_id.assert_called_once_with(profile, None, True)


@pytest.mark.asyncio
async def test_mediation_record_if_id_with_id(
    profile: Profile, route_manager: RouteManager
):
    mediation_record = MediationRecord(
        mediation_id="test-mediation-id", state=MediationRecord.STATE_GRANTED
    )
    with mock.patch.object(
        MediationRecord,
        "retrieve_by_id",
        mock.CoroutineMock(return_value=mediation_record),
    ) as mock_retrieve_by_id:
        actual = await route_manager.mediation_record_if_id(
            profile, mediation_id=mediation_record.mediation_id
        )
        assert mediation_record == actual
        mock_retrieve_by_id.assert_called_once()


@pytest.mark.asyncio
async def test_mediation_record_if_id_with_id_bad_state(
    profile: Profile, route_manager: RouteManager
):
    mediation_record = MediationRecord(
        mediation_id="test-mediation-id", state=MediationRecord.STATE_DENIED
    )
    with mock.patch.object(
        MediationRecord,
        "retrieve_by_id",
        mock.CoroutineMock(return_value=mediation_record),
    ):
        with pytest.raises(RouteManagerError):
            await route_manager.mediation_record_if_id(
                profile, mediation_id=mediation_record.mediation_id
            )


@pytest.mark.asyncio
async def test_mediation_record_if_id_with_id_and_default(
    profile: Profile, route_manager: RouteManager
):
    mediation_record = MediationRecord(
        mediation_id="test-mediation-id", state=MediationRecord.STATE_GRANTED
    )
    with (
        mock.patch.object(
            MediationRecord,
            "retrieve_by_id",
            mock.CoroutineMock(return_value=mediation_record),
        ) as mock_retrieve_by_id,
        mock.patch.object(
            MediationManager, "get_default_mediator", mock.CoroutineMock()
        ) as mock_get_default_mediator,
    ):
        actual = await route_manager.mediation_record_if_id(
            profile, mediation_id=mediation_record.mediation_id, or_default=True
        )
        assert mediation_record == actual
        mock_retrieve_by_id.assert_called_once()
        mock_get_default_mediator.assert_not_called()


@pytest.mark.asyncio
async def test_mediation_record_if_id_without_id_and_default(
    profile: Profile,
    route_manager: RouteManager,
):
    mediation_record = MediationRecord(
        mediation_id="test-mediation-id", state=MediationRecord.STATE_GRANTED
    )
    with (
        mock.patch.object(
            MediationRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_retrieve_by_id,
        mock.patch.object(
            MediationManager,
            "get_default_mediator",
            mock.CoroutineMock(return_value=mediation_record),
        ) as mock_get_default_mediator,
    ):
        actual = await route_manager.mediation_record_if_id(
            profile, mediation_id=None, or_default=True
        )
        assert mediation_record == actual
        mock_retrieve_by_id.assert_not_called()
        mock_get_default_mediator.assert_called_once()


@pytest.mark.asyncio
async def test_mediation_record_if_id_without_id_and_no_default(
    profile: Profile,
    route_manager: RouteManager,
):
    with (
        mock.patch.object(
            MediationRecord, "retrieve_by_id", mock.CoroutineMock(return_value=None)
        ) as mock_retrieve_by_id,
        mock.patch.object(
            MediationManager,
            "get_default_mediator",
            mock.CoroutineMock(return_value=None),
        ) as mock_get_default_mediator,
    ):
        assert (
            await route_manager.mediation_record_if_id(
                profile, mediation_id=None, or_default=True
            )
            is None
        )
        mock_retrieve_by_id.assert_not_called()
        mock_get_default_mediator.assert_called_once()


@pytest.mark.asyncio
async def test_route_connection_as_invitee(
    profile: Profile, route_manager: RouteManager, conn_record: ConnRecord
):
    mediation_record = MediationRecord(mediation_id="test-mediation-id")
    mock_did_info = mock.MagicMock(DIDInfo)
    with mock.patch.object(
        route_manager,
        "get_or_create_my_did",
        mock.CoroutineMock(return_value=mock_did_info),
    ):
        await route_manager.route_connection_as_invitee(
            profile, conn_record, [mediation_record]
        )
        route_manager._route_for_key.assert_called_once_with(
            profile, mock_did_info.verkey, mediation_record, skip_if_exists=True
        )


@pytest.mark.asyncio
async def test_route_connection_as_inviter(
    profile: Profile, route_manager: RouteManager, conn_record: ConnRecord
):
    mediation_record = MediationRecord(mediation_id="test-mediation-id")
    mock_did_info = mock.MagicMock(DIDInfo)
    conn_record.invitation_key = "test-invitation-key"
    with mock.patch.object(
        route_manager,
        "get_or_create_my_did",
        mock.CoroutineMock(return_value=mock_did_info),
    ):
        await route_manager.route_connection_as_inviter(
            profile, conn_record, [mediation_record]
        )
        route_manager._route_for_key.assert_called_once_with(
            profile,
            mock_did_info.verkey,
            mediation_record,
            replace_key="test-invitation-key",
            skip_if_exists=True,
        )


@pytest.mark.asyncio
async def test_route_connection_state_inviter_replace_key_none(
    profile: Profile, route_manager: RouteManager, conn_record: ConnRecord
):
    mediation_record = MediationRecord(mediation_id="test-mediation-id")
    mock_did_info = mock.MagicMock(DIDInfo)
    conn_record.invitation_key = TEST_RECORD_VERKEY

    with (
        mock.patch.object(
            route_manager,
            "get_or_create_my_did",
            mock.CoroutineMock(return_value=mock_did_info),
        ),
        mock.patch.object(
            AskarWallet,
            "get_public_did",
            mock.CoroutineMock(
                return_value=DIDInfo(
                    TEST_RECORD_DID,
                    TEST_RECORD_VERKEY,
                    None,
                    method=SOV,
                    key_type=ED25519,
                )
            ),
        ),
    ):
        await route_manager.route_connection_as_inviter(
            profile, conn_record, [mediation_record]
        )
        route_manager._route_for_key.assert_called_once_with(
            profile,
            mock_did_info.verkey,
            mediation_record,
            replace_key=None,
            skip_if_exists=True,
        )


@pytest.mark.asyncio
async def test_route_connection_state_invitee(
    profile: Profile, route_manager: RouteManager, conn_record: ConnRecord
):
    mediation_record = MediationRecord(mediation_id="test-mediation-id")
    conn_record.state = "invitation"
    conn_record.their_role = "responder"
    with (
        mock.patch.object(
            route_manager, "route_connection_as_invitee", mock.CoroutineMock()
        ) as mock_route_connection_as_invitee,
        mock.patch.object(
            route_manager, "route_connection_as_inviter", mock.CoroutineMock()
        ) as mock_route_connection_as_inviter,
    ):
        await route_manager.route_connection(profile, conn_record, [mediation_record])
        mock_route_connection_as_invitee.assert_called_once()
        mock_route_connection_as_inviter.assert_not_called()


@pytest.mark.asyncio
async def test_route_connection_state_inviter(
    profile: Profile, route_manager: RouteManager, conn_record: ConnRecord
):
    mediation_record = MediationRecord(mediation_id="test-mediation-id")
    conn_record.state = "request"
    conn_record.their_role = "requester"
    with (
        mock.patch.object(
            route_manager, "route_connection_as_invitee", mock.CoroutineMock()
        ) as mock_route_connection_as_invitee,
        mock.patch.object(
            route_manager, "route_connection_as_inviter", mock.CoroutineMock()
        ) as mock_route_connection_as_inviter,
    ):
        await route_manager.route_connection(profile, conn_record, [mediation_record])
        mock_route_connection_as_inviter.assert_called_once()
        mock_route_connection_as_invitee.assert_not_called()


@pytest.mark.asyncio
async def test_route_connection_state_other(
    profile: Profile, route_manager: RouteManager, conn_record: ConnRecord
):
    mediation_record = MediationRecord(mediation_id="test-mediation-id")
    conn_record.state = "response"
    conn_record.their_role = "requester"
    assert (
        await route_manager.route_connection(profile, conn_record, [mediation_record])
        is None
    )


@pytest.mark.asyncio
async def test_route_invitation_with_key(
    profile: Profile, route_manager: RouteManager, conn_record: ConnRecord
):
    mediation_record = MediationRecord(mediation_id="test-mediation-id")
    conn_record.invitation_key = "test-invitation-key"
    with mock.patch.object(
        route_manager, "save_mediator_for_connection", mock.CoroutineMock()
    ):
        await route_manager.route_invitation(profile, conn_record, mediation_record)
        route_manager._route_for_key.assert_called_once()


@pytest.mark.asyncio
async def test_route_invitation_without_key(
    profile: Profile, route_manager: RouteManager, conn_record: ConnRecord
):
    mediation_record = MediationRecord(mediation_id="test-mediation-id")
    with mock.patch.object(
        route_manager, "save_mediator_for_connection", mock.CoroutineMock()
    ):
        with pytest.raises(ValueError):
            await route_manager.route_invitation(profile, conn_record, mediation_record)
        route_manager._route_for_key.assert_not_called()


@pytest.mark.asyncio
async def test_route_public_did(profile: Profile, route_manager: RouteManager):
    await route_manager.route_public_did(profile, "test-verkey")
    route_manager._route_for_key.assert_called_once_with(
        profile, "test-verkey", skip_if_exists=True
    )


@pytest.mark.asyncio
async def test_route_verkey(profile: Profile, route_manager: RouteManager):
    await route_manager.route_verkey(profile, "test-verkey")
    route_manager._route_for_key.assert_called_once_with(
        profile, "test-verkey", None, skip_if_exists=True
    )


@pytest.mark.asyncio
async def test_route_static(
    profile: Profile, route_manager: RouteManager, conn_record: ConnRecord
):
    mediation_record = MediationRecord(mediation_id="test-mediation-id")
    mock_did_info = mock.MagicMock(DIDInfo)
    conn_record.invitation_key = "test-invitation-key"
    with mock.patch.object(
        route_manager,
        "get_or_create_my_did",
        mock.CoroutineMock(return_value=mock_did_info),
    ):
        await route_manager.route_static(profile, conn_record, mediation_record)
        route_manager._route_for_key.assert_called_once_with(
            profile,
            mock_did_info.verkey,
            mediation_record,
            skip_if_exists=True,
        )


@pytest.mark.asyncio
async def test_save_mediator_for_connection_record(
    profile: Profile,
    route_manager: RouteManager,
    conn_record: ConnRecord,
):
    mediation_record = MediationRecord(mediation_id="test-mediation-id")
    session = mock.MagicMock()
    profile.session = mock.MagicMock(return_value=session)
    session.__aenter__ = mock.CoroutineMock(return_value=session)
    session.__aexit__ = mock.CoroutineMock()
    with mock.patch.object(
        MediationRecord, "retrieve_by_id", mock.CoroutineMock()
    ) as mock_retrieve_by_id:
        await route_manager.save_mediator_for_connection(
            profile, conn_record, mediation_record
        )
        mock_retrieve_by_id.assert_not_called()
        conn_record.metadata_set.assert_called_once_with(
            session,
            MediationManager.METADATA_KEY,
            {MediationManager.METADATA_ID: mediation_record.mediation_id},
        )


@pytest.mark.asyncio
async def test_save_mediator_for_connection_id(
    profile: Profile,
    route_manager: RouteManager,
    conn_record: ConnRecord,
):
    mediation_record = MediationRecord(mediation_id="test-mediation-id")
    session = mock.MagicMock()
    profile.session = mock.MagicMock(return_value=session)
    session.__aenter__ = mock.CoroutineMock(return_value=session)
    session.__aexit__ = mock.CoroutineMock()
    with mock.patch.object(
        MediationRecord,
        "retrieve_by_id",
        mock.CoroutineMock(return_value=mediation_record),
    ) as mock_retrieve_by_id:
        await route_manager.save_mediator_for_connection(
            profile, conn_record, mediation_id=mediation_record.mediation_id
        )
        mock_retrieve_by_id.assert_called_once()
        conn_record.metadata_set.assert_called_once_with(
            session,
            MediationManager.METADATA_KEY,
            {MediationManager.METADATA_ID: mediation_record.mediation_id},
        )


@pytest.mark.asyncio
async def test_save_mediator_for_connection_no_mediator(
    profile: Profile,
    route_manager: RouteManager,
    conn_record: ConnRecord,
):
    with mock.patch.object(
        MediationRecord, "retrieve_by_id", mock.CoroutineMock()
    ) as mock_retrieve_by_id:
        await route_manager.save_mediator_for_connection(profile, conn_record)
        mock_retrieve_by_id.assert_not_called()
        conn_record.metadata_set.assert_not_called()


@pytest.mark.asyncio
async def test_connection_from_recipient_key_invite(
    profile: Profile, route_manager: RouteManager, conn_record: ConnRecord
):
    with mock.patch.object(
        ConnRecord,
        "retrieve_by_tag_filter",
        mock.CoroutineMock(return_value=conn_record),
    ):
        result = await route_manager.connection_from_recipient_key(profile, TEST_VERKEY)
        assert conn_record == result


@pytest.mark.asyncio
async def test_connection_from_recipient_key_local_did(
    profile: Profile, route_manager: RouteManager, conn_record: ConnRecord
):
    mock_wallet = mock.MagicMock(AskarWallet, autospec=True)
    with mock.patch.object(AskarWallet, "get_local_did_for_verkey"):
        profile.context.injector.bind_instance(AskarWallet, mock_wallet)
        with mock.patch.object(
            ConnRecord, "retrieve_by_did", mock.CoroutineMock(return_value=conn_record)
        ):
            result = await route_manager.connection_from_recipient_key(
                profile, TEST_VERKEY
            )
            assert conn_record == result


@pytest.mark.asyncio
async def test_mediation_route_for_key(
    profile: Profile,
    mediation_route_manager: CoordinateMediationV1RouteManager,
    mock_responder: MockResponder,
):
    mediation_record = MediationRecord(
        mediation_id="test-mediation-id", connection_id="test-mediator-conn-id"
    )
    keylist_update = await mediation_route_manager._route_for_key(
        profile,
        TEST_VERKEY,
        mediation_record,
        skip_if_exists=False,
        replace_key=None,
    )
    assert keylist_update
    assert keylist_update.serialize()["updates"] == [
        {"action": "add", "recipient_key": TEST_VERKEY}
    ]
    assert mock_responder.messages
    assert (
        keylist_update,
        {"connection_id": "test-mediator-conn-id"},
    ) == mock_responder.messages[0]


@pytest.mark.asyncio
async def test_mediation_route_for_key_skip_if_exists_and_exists(
    profile: Profile,
    mediation_route_manager: CoordinateMediationV1RouteManager,
    mock_responder: MockResponder,
):
    mediation_record = MediationRecord(
        mediation_id="test-mediation-id", connection_id="test-mediator-conn-id"
    )
    with mock.patch.object(
        RouteRecord, "retrieve_by_recipient_key", mock.CoroutineMock()
    ):
        keylist_update = await mediation_route_manager._route_for_key(
            profile,
            TEST_VERKEY,
            mediation_record,
            skip_if_exists=True,
            replace_key=None,
        )
    assert keylist_update is None
    assert not mock_responder.messages


@pytest.mark.asyncio
async def test_mediation_route_for_key_skip_if_exists_and_absent(
    profile: Profile,
    mediation_route_manager: CoordinateMediationV1RouteManager,
    mock_responder: MockResponder,
):
    mediation_record = MediationRecord(
        mediation_id="test-mediation-id", connection_id="test-mediator-conn-id"
    )
    with mock.patch.object(
        RouteRecord,
        "retrieve_by_recipient_key",
        mock.CoroutineMock(side_effect=StorageNotFoundError),
    ):
        keylist_update = await mediation_route_manager._route_for_key(
            profile,
            TEST_VERKEY,
            mediation_record,
            skip_if_exists=True,
            replace_key=None,
        )
    assert keylist_update
    assert keylist_update.serialize()["updates"] == [
        {"action": "add", "recipient_key": TEST_VERKEY}
    ]
    assert mock_responder.messages
    assert (
        keylist_update,
        {"connection_id": "test-mediator-conn-id"},
    ) == mock_responder.messages[0]


@pytest.mark.asyncio
async def test_mediation_route_for_key_replace_key(
    profile: Profile,
    mediation_route_manager: CoordinateMediationV1RouteManager,
    mock_responder: MockResponder,
):
    mediation_record = MediationRecord(
        mediation_id="test-mediation-id", connection_id="test-mediator-conn-id"
    )
    keylist_update = await mediation_route_manager._route_for_key(
        profile,
        TEST_VERKEY,
        mediation_record,
        skip_if_exists=False,
        replace_key=TEST_ROUTE_VERKEY,
    )
    assert keylist_update
    assert keylist_update.serialize()["updates"] == [
        {"action": "add", "recipient_key": TEST_VERKEY},
        {"action": "remove", "recipient_key": TEST_ROUTE_VERKEY},
    ]
    assert mock_responder.messages
    assert (
        keylist_update,
        {"connection_id": "test-mediator-conn-id"},
    ) == mock_responder.messages[0]


@pytest.mark.asyncio
async def test_mediation_route_for_key_no_mediator(
    profile: Profile,
    mediation_route_manager: CoordinateMediationV1RouteManager,
):
    assert (
        await mediation_route_manager._route_for_key(
            profile,
            TEST_VERKEY,
            None,
            skip_if_exists=True,
            replace_key=TEST_ROUTE_VERKEY,
        )
        is None
    )


@pytest.mark.asyncio
async def test_mediation_routing_info_with_mediator(
    profile: Profile,
    mediation_route_manager: CoordinateMediationV1RouteManager,
):
    mediation_record = MediationRecord(
        mediation_id="test-mediation-id",
        connection_id="test-mediator-conn-id",
        routing_keys=[TEST_ROUTE_VERKEY_REF],
        endpoint="http://mediator.example.com",
    )
    keys, endpoint = await mediation_route_manager.routing_info(profile, mediation_record)
    assert keys == mediation_record.routing_keys
    assert endpoint == mediation_record.endpoint


@pytest.mark.asyncio
async def test_mediation_routing_info_no_mediator(
    profile: Profile,
    mediation_route_manager: CoordinateMediationV1RouteManager,
):
    keys, endpoint = await mediation_route_manager.routing_info(profile, None)
    assert keys is None
    assert endpoint is None



"""Environment utility methods."""

from os import getenv
from pathlib import Path


def storage_path(*subpaths, create: bool = False) -> Path:
    """Get the default aca-py home directory."""
    custom = getenv("ACAPY_HOME")
    if custom:
        path = Path(custom)
    else:
        path = Path.home().joinpath(".acapy_agent")
    if subpaths:
        path = path.joinpath(*subpaths)
    if create:
        path.mkdir(parents=True, exist_ok=True)
    return path



"""Represents an OOB connection reuse problem report message."""

import logging
from enum import Enum
from typing import Optional, Text

from marshmallow import EXCLUDE, ValidationError, fields, pre_dump, validates_schema

from ....problem_report.v1_0.message import ProblemReport, ProblemReportSchema
from ..message_types import DEFAULT_VERSION, PROBLEM_REPORT, PROTOCOL_PACKAGE

HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers.problem_report_handler.OOBProblemReportMessageHandler"
)

LOGGER = logging.getLogger(__name__)


class ProblemReportReason(Enum):
    """Supported reason codes."""

    NO_EXISTING_CONNECTION = "no_existing_connection"
    EXISTING_CONNECTION_NOT_ACTIVE = "existing_connection_not_active"


class OOBProblemReport(ProblemReport):
    """Base class representing an OOB connection reuse problem report message."""

    class Meta:
        """OOB connection reuse problem report metadata."""

        handler_class = HANDLER_CLASS
        message_type = PROBLEM_REPORT
        schema_class = "OOBProblemReportSchema"

    def __init__(
        self,
        version: str = DEFAULT_VERSION,
        msg_type: Optional[Text] = None,
        *args,
        **kwargs,
    ):
        """Initialize a ProblemReport message instance."""
        super().__init__(_type=msg_type, _version=version, *args, **kwargs)


class OOBProblemReportSchema(ProblemReportSchema):
    """Schema for ProblemReport base class."""

    class Meta:
        """Metadata for problem report schema."""

        model_class = OOBProblemReport
        unknown = EXCLUDE

    _type = fields.Str(
        data_key="@type",
        required=False,
        metadata={
            "description": "Message type",
            "example": "https://didcomm.org/my-family/1.0/my-message-type",
        },
    )

    @pre_dump
    def check_thread_deco(self, obj, **kwargs):
        """Thread decorator, and its thid and pthid, are mandatory."""

        if not obj._decorators.to_dict().get("~thread", {}).keys() >= {"thid", "pthid"}:
            raise ValidationError("Missing required field(s) in thread decorator")

        return obj

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields."""

        if not data.get("description", {}).get("code", ""):
            raise ValidationError("Value for description.code must be present")
        elif data.get("description", {}).get("code", "") not in [
            prr.value for prr in ProblemReportReason
        ]:
            locales = list(data.get("description").keys())
            locales.remove("code")
            LOGGER.warning(
                "Unexpected error code received.\n"
                f"Code: {data.get('description').get('code')}, "
                f"Description: {data.get('description').get(locales[0])}"
            )



"""Ledger related errors."""

from typing import Generic, Optional, TypeVar

from ..core.error import BaseError


class LedgerError(BaseError):
    """Base class for ledger errors."""


class LedgerConfigError(LedgerError):
    """Base class for ledger configuration errors."""


class BadLedgerRequestError(LedgerError):
    """The current request cannot proceed."""


class ClosedPoolError(LedgerError):
    """Indy pool is closed."""


class LedgerTransactionError(LedgerError):
    """The ledger rejected the transaction."""


T = TypeVar("T")


class LedgerObjectAlreadyExistsError(LedgerError, Generic[T]):
    """Raised when a ledger object already existed."""

    def __init__(
        self,
        message: str,
        obj_id: str,
        obj: Optional[T] = None,
        *args,
        **kwargs,
    ):
        """Initialize an instance.

        Args:
            message: Human readable message text
            obj_id: ledger object id
            obj: ledger object
            args: additional positional arguments
            kwargs: additional keyword arguments

        """
        super().__init__(message, obj_id, obj, *args, **kwargs)
        self._message = message
        self.obj_id = obj_id
        self.obj = obj

    @property
    def message(self):
        """Error message."""
        return f"{self._message}: {self.obj_id}, {self.obj}"



"""Upgrade command for handling breaking changes when updating ACA-PY versions."""

import asyncio
import json
import logging
import os
from enum import Enum
from typing import Any, Callable, List, Mapping, Optional, Sequence, Tuple, Union

import yaml
from configargparse import ArgumentParser
from packaging import version as package_version

from ..config import argparse as arg
from ..config.base import BaseError, BaseSettings
from ..config.default_context import DefaultContextBuilder
from ..config.injection_context import InjectionContext
from ..config.util import common_config
from ..config.wallet import wallet_config
from ..core.profile import Profile, ProfileSession
from ..messaging.models.base import BaseModelError
from ..messaging.models.base_record import BaseRecord, RecordType
from ..revocation.models.issuer_rev_reg_record import IssuerRevRegRecord
from ..storage.base import BaseStorage, BaseStorageSearch
from ..storage.error import StorageNotFoundError
from ..storage.record import StorageRecord
from ..utils.classloader import ClassLoader, ClassNotFoundError
from ..version import RECORD_TYPE_ACAPY_VERSION, __version__
from ..wallet.models.wallet_record import WalletRecord
from . import PROG

DEFAULT_UPGRADE_CONFIG_FILE_NAME = "default_version_upgrade_config.yml"
LOGGER = logging.getLogger(__name__)
BATCH_SIZE = 25


class ExplicitUpgradeOption(Enum):
    """Supported explicit upgrade codes."""

    ERROR_AND_STOP = "critical"
    LOG_AND_PROCEED = "warning"

    @classmethod
    def get(cls, code: str) -> "ExplicitUpgradeOption":
        """Get ExplicitUpgradeOption for code."""
        for option in ExplicitUpgradeOption:
            if code == option.value:
                return option
        return None


class UpgradeError(BaseError):
    """Base exception for upgrade related errors."""


class VersionUpgradeConfig:
    """Handle ACA-Py version upgrade config."""

    def __init__(self, config_path: Optional[str] = None):
        """Initialize config for use during upgrade process."""
        self.function_map_config = UPGRADE_EXISTING_RECORDS_FUNCTION_MAPPING
        self.upgrade_configs = {}
        if config_path:
            self.setup_version_upgrade_config(config_path)
        else:
            self.setup_version_upgrade_config(
                os.path.join(
                    os.path.dirname(os.path.realpath(__file__)),
                    DEFAULT_UPGRADE_CONFIG_FILE_NAME,
                )
            )

    def setup_version_upgrade_config(self, path: str):
        """Set ups config dict from the provided YML file."""
        with open(path, "r") as stream:
            config_dict = yaml.safe_load(stream)
            tagged_config_dict = {}
            for config_id, provided_config in config_dict.items():
                recs_list = []
                tagged_config_dict[config_id] = {}
                if "resave_records" in provided_config:
                    if provided_config.get("resave_records").get("base_record_path"):
                        recs_list = recs_list + provided_config.get("resave_records").get(
                            "base_record_path"
                        )
                    if provided_config.get("resave_records").get("base_exch_record_path"):
                        recs_list = recs_list + provided_config.get("resave_records").get(
                            "base_exch_record_path"
                        )
                tagged_config_dict[config_id]["resave_records"] = recs_list
                config_key_set = set(provided_config.keys())
                try:
                    config_key_set.remove("resave_records")
                except KeyError:
                    pass
                if "explicit_upgrade" in provided_config:
                    tagged_config_dict[config_id]["explicit_upgrade"] = (
                        provided_config.get("explicit_upgrade")
                    )
                try:
                    config_key_set.remove("explicit_upgrade")
                except KeyError:
                    pass
                for executable in config_key_set:
                    tagged_config_dict[config_id][executable] = (
                        provided_config.get(executable) or False
                    )
            if tagged_config_dict == {}:
                raise UpgradeError(f"No version configs found in {path}")
            self.upgrade_configs = tagged_config_dict

    def get_callable(self, executable: str) -> Optional[Callable]:
        """Return callable function for executable name."""
        if executable in self.function_map_config:
            return self.function_map_config.get(executable)
        else:
            return None


def init_argument_parser(parser: ArgumentParser):
    """Initialize an argument parser with the module's arguments."""
    return arg.load_argument_groups(parser, *arg.group.get_registered(arg.CAT_UPGRADE))


def explicit_upgrade_required_check(
    to_apply_version_list: List,
    upgrade_config: dict,
) -> Tuple[bool, List, Optional[str]]:
    """Check if explicit upgrade is required."""
    to_skip_versions = []
    for version in to_apply_version_list:
        if "explicit_upgrade" in upgrade_config.get(version):
            exp_upg_flag = upgrade_config.get(version).get("explicit_upgrade")
            if (
                ExplicitUpgradeOption.get(exp_upg_flag)
                is ExplicitUpgradeOption.ERROR_AND_STOP
            ):
                return True, [], version
            elif (
                ExplicitUpgradeOption.get(exp_upg_flag)
                is ExplicitUpgradeOption.LOG_AND_PROCEED
            ):
                to_skip_versions.append(version)
    return False, to_skip_versions, None


def get_upgrade_version_list(
    from_version: str,
    sorted_version_list: Optional[List] = None,
    config_path: Optional[str] = None,
) -> List:
    """Get available versions from the upgrade config."""
    if not sorted_version_list:
        version_upgrade_config_inst = VersionUpgradeConfig(config_path)
        upgrade_configs = version_upgrade_config_inst.upgrade_configs
        tags_found_in_config = upgrade_configs.keys()
        version_found_in_config, _ = _get_version_and_name_tags(
            list(tags_found_in_config)
        )
        sorted_version_list = sorted(
            version_found_in_config, key=lambda x: package_version.parse(x)
        )

    version_list = []
    for version in sorted_version_list:
        if package_version.parse(version) >= package_version.parse(from_version):
            version_list.append(version)
    return version_list


async def add_version_record(profile: Profile, version: str):
    """Add an acapy_version storage record for provided version."""
    async with profile.session() as session:
        storage = session.inject(BaseStorage)
        try:
            version_storage_record = await storage.find_record(
                type_filter=RECORD_TYPE_ACAPY_VERSION, tag_query={}
            )
        except StorageNotFoundError:
            version_storage_record = None
        if not version_storage_record:
            await storage.add_record(
                StorageRecord(
                    RECORD_TYPE_ACAPY_VERSION,
                    version,
                )
            )
        else:
            await storage.update_record(version_storage_record, version, {})
    LOGGER.info(f"{RECORD_TYPE_ACAPY_VERSION} storage record set to {version}")


def _get_version_and_name_tags(tags_found_in_config: List) -> Tuple[List, List]:
    """Get version and named tag key lists from config."""
    version_found_in_config = []
    named_tag_found_in_config = []
    for tag in tags_found_in_config:
        try:
            package_version.parse(tag)
            version_found_in_config.append(tag)
        except package_version.InvalidVersion:
            named_tag_found_in_config.append(tag)
    return version_found_in_config, named_tag_found_in_config


def _perform_upgrade(
    upgrade_config: dict,
    resave_record_path_sets: set,
    executables_call_set: set,
    tag: str,
) -> Tuple[set, set]:
    """Update and return resave record path and executables call sets."""
    LOGGER.info(f"Running upgrade process for {tag}")
    # Step 1 re-saving all BaseRecord and BaseExchangeRecord
    if "resave_records" in upgrade_config:
        resave_record_paths = upgrade_config.get("resave_records")
        for record_path in resave_record_paths:
            resave_record_path_sets.add(record_path)

    # Step 2 Update existing records, if required
    config_key_set = set(upgrade_config.keys())
    try:
        config_key_set.remove("resave_records")
    except KeyError:
        pass
    for callable_name in list(config_key_set):
        if upgrade_config.get(callable_name) is False:
            continue
        executables_call_set.add(callable_name)
    return resave_record_path_sets, executables_call_set


def get_webhook_urls(
    base_context: InjectionContext,
    wallet_record: WalletRecord,
) -> list:
    """Get the webhook urls according to dispatch_type."""
    wallet_id = wallet_record.wallet_id
    dispatch_type = wallet_record.wallet_dispatch_type
    subwallet_webhook_urls = wallet_record.wallet_webhook_urls or []
    base_webhook_urls = base_context.settings.get("admin.webhook_urls", [])

    if dispatch_type == "both":
        webhook_urls = list(set(base_webhook_urls) | set(subwallet_webhook_urls))
        if not webhook_urls:
            LOGGER.warning(
                "No webhook URLs in context configuration "
                f"nor wallet record {wallet_id}, but wallet record "
                f"configures dispatch type {dispatch_type}"
            )
    elif dispatch_type == "default":
        webhook_urls = subwallet_webhook_urls
        if not webhook_urls:
            LOGGER.warning(
                f"No webhook URLs in nor wallet record {wallet_id}, but "
                f"wallet record configures dispatch type {dispatch_type}"
            )
    else:
        webhook_urls = base_webhook_urls
    return webhook_urls


async def get_wallet_profile(
    base_context: InjectionContext,
    wallet_record: WalletRecord,
    extra_settings: Optional[dict] = None,
) -> Profile:
    """Get profile for a wallet record."""
    extra_settings = extra_settings or {}
    context = base_context.copy()
    reset_settings = {
        "wallet.recreate": False,
        "wallet.seed": None,
        "wallet.rekey": None,
        "wallet.name": None,
        "wallet.type": None,
        "mediation.open": None,
        "mediation.invite": None,
        "mediation.default_id": None,
        "mediation.clear": None,
    }
    extra_settings["admin.webhook_urls"] = get_webhook_urls(base_context, wallet_record)

    context.settings = (
        context.settings.extend(reset_settings)
        .extend(wallet_record.settings)
        .extend(extra_settings)
    )

    profile, _ = await wallet_config(context, provision=False)
    return profile


async def upgrade(
    settings: Optional[Union[Mapping[str, Any], BaseSettings]] = None,
    profile: Optional[Profile] = None,
):
    """Invoke upgradation process for each applicable profile."""
    profiles_to_upgrade = []
    if settings:
        batch_size = settings.get("upgrade.page_size", BATCH_SIZE)
    else:
        batch_size = BATCH_SIZE
    if profile and (settings or settings == {}):
        raise UpgradeError("upgrade requires either profile or settings, not both.")
    if profile:
        root_profile = profile
        settings = profile.settings
    else:
        context_builder = DefaultContextBuilder(settings)
        context = await context_builder.build_context()
        root_profile, _ = await wallet_config(context)
    profiles_to_upgrade.append(root_profile)
    base_storage_search_inst = root_profile.inject(BaseStorageSearch)
    if "upgrade.upgrade_all_subwallets" in settings and settings.get(
        "upgrade.upgrade_all_subwallets"
    ):
        search_session = base_storage_search_inst.search_records(
            type_filter=WalletRecord.RECORD_TYPE, page_size=batch_size
        )
        while search_session._done is False:
            wallet_storage_records = await search_session.fetch()
            for wallet_storage_record in wallet_storage_records:
                wallet_record = WalletRecord.from_storage(
                    wallet_storage_record.id,
                    json.loads(wallet_storage_record.value),
                )
                wallet_profile = await get_wallet_profile(
                    base_context=root_profile.context, wallet_record=wallet_record
                )
                profiles_to_upgrade.append(wallet_profile)
        del settings["upgrade.upgrade_all_subwallets"]
    if (
        "upgrade.upgrade_subwallets" in settings
        and len(settings.get("upgrade.upgrade_subwallets")) >= 1
    ):
        for _wallet_id in settings.get("upgrade.upgrade_subwallets"):
            async with root_profile.session() as session:
                wallet_record = await WalletRecord.retrieve_by_id(
                    session, record_id=_wallet_id
                )
            wallet_profile = await get_wallet_profile(
                base_context=root_profile.context, wallet_record=wallet_record
            )
            profiles_to_upgrade.append(wallet_profile)
        del settings["upgrade.upgrade_subwallets"]
    for _profile in profiles_to_upgrade:
        await upgrade_per_profile(profile=_profile, settings=settings)


async def upgrade_per_profile(
    profile: Profile,
    settings: Optional[Union[Mapping[str, Any], BaseSettings]] = None,
):
    """Perform upgradation steps."""
    try:
        version_upgrade_config_inst = VersionUpgradeConfig(
            settings.get("upgrade.config_path")
        )
        upgrade_from_tags = None
        force_upgrade_flag = settings.get("upgrade.force_upgrade") or False
        if force_upgrade_flag:
            upgrade_from_tags = settings.get("upgrade.named_tags")
        upgrade_configs = version_upgrade_config_inst.upgrade_configs
        upgrade_to_version = f"v{__version__}"
        tags_found_in_config = upgrade_configs.keys()
        version_found_in_config, named_tag_found_in_config = _get_version_and_name_tags(
            list(tags_found_in_config)
        )
        sorted_versions_found_in_config = sorted(
            version_found_in_config, key=lambda x: package_version.parse(x)
        )
        upgrade_from_version_storage = None
        upgrade_from_version_config = None
        upgrade_from_version = None
        async with profile.session() as session:
            storage = session.inject(BaseStorage)
            try:
                version_storage_record = await storage.find_record(
                    type_filter=RECORD_TYPE_ACAPY_VERSION, tag_query={}
                )
                upgrade_from_version_storage = version_storage_record.value
            except StorageNotFoundError:
                LOGGER.info("No ACA-Py version found in wallet storage.")
                version_storage_record = None

            if "upgrade.from_version" in settings:
                upgrade_from_version_config = settings.get("upgrade.from_version")
                LOGGER.info(
                    (
                        f"Selecting {upgrade_from_version_config} as "
                        "--from-version from the config."
                    )
                )

        if upgrade_from_version_storage and upgrade_from_version_config:
            if (
                package_version.parse(upgrade_from_version_storage)
                > package_version.parse(upgrade_from_version_config)
            ) and force_upgrade_flag:
                upgrade_from_version = upgrade_from_version_config
            else:
                upgrade_from_version = upgrade_from_version_storage
        if (
            not upgrade_from_version
            and not upgrade_from_version_storage
            and upgrade_from_version_config
        ):
            upgrade_from_version = upgrade_from_version_config
        if (
            not upgrade_from_version
            and upgrade_from_version_storage
            and not upgrade_from_version_config
        ):
            upgrade_from_version = upgrade_from_version_storage
        if not upgrade_from_version and not upgrade_from_tags:
            raise UpgradeError(
                "No upgrade from version or tags found in wallet"
                " or settings [--from-version or --named-tag]"
            )
        resave_record_path_sets = set()
        executables_call_set = set()
        to_update_flag = False
        if upgrade_from_version:
            upgrade_version_in_config = get_upgrade_version_list(
                sorted_version_list=sorted_versions_found_in_config,
                from_version=upgrade_from_version,
            )
            # Perform explicit upgrade check if the function was called during startup
            if profile:
                (
                    explicit_flag,
                    to_skip_explicit_versions,
                    explicit_upg_ver,
                ) = explicit_upgrade_required_check(
                    to_apply_version_list=upgrade_version_in_config,
                    upgrade_config=upgrade_configs,
                )
                if explicit_flag:
                    raise UpgradeError(
                        "Explicit upgrade flag with critical value found "
                        f"for {explicit_upg_ver} config. Please use ACA-Py "
                        "upgrade command to complete the process and proceed."
                    )
                if len(to_skip_explicit_versions) >= 1:
                    LOGGER.warning(
                        "Explicit upgrade flag with warning value found "
                        f"for {str(to_skip_explicit_versions)} versions. "
                        "Proceeding with ACA-Py startup. You can apply "
                        "the explicit upgrades using the ACA-Py upgrade "
                        "command later."
                    )
                    return
            if upgrade_from_version == upgrade_to_version:
                LOGGER.info(
                    (
                        f"Version {upgrade_from_version} to upgrade from and "
                        f"current version to upgrade to {upgrade_to_version} "
                        "are same. You can apply upgrade from a lower "
                        "version by running the upgrade command with "
                        f"--from-version [< {upgrade_to_version}] and "
                        "--force-upgrade"
                    )
                )
            else:
                for config_from_version in upgrade_version_in_config:
                    resave_record_path_sets, executables_call_set = _perform_upgrade(
                        upgrade_config=upgrade_configs.get(config_from_version),
                        resave_record_path_sets=resave_record_path_sets,
                        executables_call_set=executables_call_set,
                        tag=config_from_version,
                    )
        if upgrade_from_tags and len(upgrade_from_tags) >= 1:
            for named_tag in upgrade_from_tags:
                if named_tag not in named_tag_found_in_config:
                    continue
                resave_record_path_sets, executables_call_set = _perform_upgrade(
                    upgrade_config=upgrade_configs.get(named_tag),
                    resave_record_path_sets=resave_record_path_sets,
                    executables_call_set=executables_call_set,
                    tag=named_tag,
                )
        if len(resave_record_path_sets) >= 1 or len(executables_call_set) >= 1:
            to_update_flag = True
        for record_path in resave_record_path_sets:
            try:
                rec_type = ClassLoader.load_class(record_path)
            except ClassNotFoundError as err:
                raise UpgradeError(f"Unknown Record type {record_path}") from err
            if not issubclass(rec_type, BaseRecord):
                raise UpgradeError(
                    f"Only BaseRecord can be resaved, found: {str(rec_type)}"
                )
            all_records = []
            if settings:
                batch_size = settings.get("upgrade.page_size", BATCH_SIZE)
            else:
                batch_size = BATCH_SIZE
            base_storage_search_inst = profile.inject(BaseStorageSearch)
            search_session = base_storage_search_inst.search_records(
                type_filter=rec_type.RECORD_TYPE, page_size=batch_size
            )
            while search_session._done is False:
                storage_records = await search_session.fetch()
                for storage_record in storage_records:
                    _record = rec_type.from_storage(
                        storage_record.id,
                        json.loads(storage_record.value),
                    )
                    all_records.append(_record)
            async with profile.session() as session:
                for record in all_records:
                    await record.save(
                        session,
                        reason="re-saving record during the upgrade process",
                    )
                if len(all_records) == 0:
                    LOGGER.info(f"No records of {str(rec_type)} found")
                else:
                    LOGGER.info(f"All recs of {str(rec_type)} successfully re-saved")
        for callable_name in executables_call_set:
            _callable = version_upgrade_config_inst.get_callable(callable_name)
            if not _callable:
                raise UpgradeError(f"No function specified for {callable_name}")
            await _callable(profile)

        # Update storage version
        if to_update_flag:
            async with profile.session() as session:
                storage = session.inject(BaseStorage)
                if not version_storage_record:
                    await storage.add_record(
                        StorageRecord(
                            RECORD_TYPE_ACAPY_VERSION,
                            upgrade_to_version,
                        )
                    )
                else:
                    await storage.update_record(
                        version_storage_record, upgrade_to_version, {}
                    )
                LOGGER.info(
                    f"{RECORD_TYPE_ACAPY_VERSION} storage record "
                    f"set to {upgrade_to_version}"
                )
        if not profile:
            await profile.close()
    except BaseError as e:
        raise UpgradeError(f"Error during upgrade: {e}")


async def update_existing_records(profile: Profile):
    """Update existing records.

    Args:
        profile: Root profile

    """
    pass


##########################################################
# Fix for ACA-Py Issue #2485
# issuance_type attribute in IssuerRevRegRecord was removed
# in 0.5.3 version. IssuerRevRegRecord created previously
# will need
##########################################################


async def find_affected_issue_rev_reg_records(
    session: ProfileSession,
) -> Sequence[RecordType]:
    """Get IssuerRevRegRecord records with issuance_type for re-saving.

    Args:
        session: The profile session to use
    """
    storage = session.inject(BaseStorage)
    rows = await storage.find_all_records(
        IssuerRevRegRecord.RECORD_TYPE,
    )
    issue_rev_reg_records_to_update = []
    for record in rows:
        vals = json.loads(record.value)
        to_update = False
        try:
            record_id = record.id
            record_id_name = IssuerRevRegRecord.RECORD_ID_NAME
            if record_id_name in vals:
                raise ValueError(f"Duplicate {record_id_name} inputs; {vals}")
            params = dict(**vals)
            # Check for issuance_type and add record_id for later tracking
            if "issuance_type" in params:
                LOGGER.info(
                    f"IssuerRevRegRecord {record_id} tagged for fixing issuance_type."
                )
                del params["issuance_type"]
                to_update = True
            params[record_id_name] = record_id
            if to_update:
                issue_rev_reg_records_to_update.append(IssuerRevRegRecord(**params))
        except BaseModelError as err:
            raise BaseModelError(f"{err}, for record id {record.id}")
    return issue_rev_reg_records_to_update


async def fix_issue_rev_reg_records(profile: Profile):
    """Update IssuerRevRegRecord records.

    Args:
        profile: Root profile

    """
    async with profile.session() as session:
        issue_rev_reg_records = await find_affected_issue_rev_reg_records(session)
        for record in issue_rev_reg_records:
            await record.save(
                session,
                reason="re-saving issue_rev_reg record without issuance type",
            )


def execute(argv: Sequence[str] = None):
    """Entrypoint."""
    parser = arg.create_argument_parser(prog=PROG)
    parser.prog += " upgrade"
    get_settings = init_argument_parser(parser)
    args = parser.parse_args(argv)
    settings = get_settings(args)
    common_config(settings)
    loop = asyncio.get_event_loop()
    loop.run_until_complete(upgrade(settings=settings))


def main():
    """Execute the main line."""
    if __name__ == "__main__":
        execute()


UPGRADE_EXISTING_RECORDS_FUNCTION_MAPPING = {
    "update_existing_records": update_existing_records,
    "fix_issue_rev_reg_records": fix_issue_rev_reg_records,
}

main()



from unittest import IsolatedAsyncioTestCase

from ...config.base import InjectionError
from ...config.injection_context import InjectionContext
from ...utils.testing import create_test_profile
from ..manager_provider import MultitenantManagerProvider


class TestProfileManagerProvider(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.profile = await create_test_profile()

    async def test_provide_manager(self):
        provider = MultitenantManagerProvider(self.profile)
        context = InjectionContext()

        self.assertEqual(
            provider.provide(context.settings, context.injector).__class__.__name__,
            "MultitenantManager",
        )

    async def test_provide_askar_profile_manager(self):
        provider = MultitenantManagerProvider(self.profile)
        context = InjectionContext()
        context.settings["multitenant.wallet_type"] = "single-wallet-askar"

        self.assertEqual(
            provider.provide(context.settings, context.injector).__class__.__name__,
            "SingleWalletAskarMultitenantManager",
        )

    async def test_invalid_manager_type(self):
        provider = MultitenantManagerProvider(self.profile)
        context = InjectionContext()
        context.settings["multitenant.wallet_type"] = "not-valid"

        with self.assertRaises(InjectionError):
            provider.provide(context.settings, context.injector)
